"""Implementation of an approximation of the likelihood function of a mixture of Gaussian and multiplicative noises with censorship with a lower limit.
"""
from typing import List, Optional, Union

import numba
import numpy as np
import pandas as pd
from scipy.special import log_ndtr

from beetroots.modelling.likelihoods import utils
from beetroots.modelling.likelihoods.abstract_likelihood import Likelihood


@numba.njit()
def cost_au(
    y: np.ndarray, f_Theta: np.ndarray, m_a: np.ndarray, s_a: np.ndarray
) -> np.ndarray:
    nll_au = 0.5 * ((y - f_Theta - m_a) / s_a) ** 2 + np.log(s_a)
    return nll_au


@numba.njit()
def cost_mu(
    log_y: np.ndarray, log_f_Theta: np.ndarray, m_m: np.ndarray, s_m: np.ndarray
) -> np.ndarray:
    nll_mu = 0.5 * ((log_y - log_f_Theta - m_m) / s_m) ** 2 + np.log(s_m)
    # nll_mu = np.where(np.isnan(log_y) | np.isinf(log_y), 10.0 ** 8, nll_mu)

    # shape = nll_mu.shape
    # nll_mu = nll_mu.ravel()
    # nll_mu[np.isnan(nll_mu)] = 0
    # nll_mu = nll_mu.reshape(shape)
    return nll_mu


@numba.njit()
def gradient_cost_au(
    y: np.ndarray,
    grad_f_Theta: np.ndarray,
    f_Theta: np.ndarray,
    grad_m_a: np.ndarray,
    m_a: np.ndarray,
    grad_s_a2: np.ndarray,
    s_a2: np.ndarray,
) -> np.ndarray:
    u_1 = (f_Theta - m_a - y) / s_a2**2
    assert u_1.shape == f_Theta.shape  # (N, L)

    u_2 = (
        2 * np.expand_dims(s_a2, axis=1) * (grad_f_Theta + grad_m_a)
        - np.expand_dims(f_Theta - m_a - y, axis=1) * grad_s_a2
    )
    assert u_2.shape == grad_f_Theta.shape  # (N, D, L)

    g_au = 0.5 * grad_s_a2 / np.expand_dims(s_a2, axis=1)
    g_au += 0.5 * np.expand_dims(u_1, axis=1) * u_2

    assert g_au.shape == grad_f_Theta.shape
    return g_au


@numba.njit()
def gradient_cost_mu(
    log_y: np.ndarray,
    grad_log_f_Theta: np.ndarray,
    log_f_Theta: np.ndarray,
    grad_m_m: np.ndarray,
    m_m: np.ndarray,
    grad_s_m2: np.ndarray,
    s_m2: np.ndarray,
) -> np.ndarray:

    u_1 = (log_f_Theta - m_m - log_y) / s_m2**2
    assert u_1.shape == log_f_Theta.shape  # (N, L)

    u_2 = (
        2 * np.expand_dims(s_m2, axis=1) * (grad_log_f_Theta + grad_m_m)
        - np.expand_dims(log_f_Theta - m_m - log_y, axis=1) * grad_s_m2
    )
    assert u_2.shape == grad_log_f_Theta.shape  # (N, D, L)

    g_mu = 0.5 * grad_s_m2 / np.expand_dims(s_m2, axis=1)
    g_mu += np.where(
        np.expand_dims(np.isnan(log_y), axis=1) + np.zeros_like(grad_log_f_Theta),
        # fake output
        np.zeros_like(grad_log_f_Theta),
        # true grad
        0.5 * np.expand_dims(u_1, axis=1) * u_2,
    )

    assert g_mu.shape == grad_log_f_Theta.shape
    return g_mu


class MixingModelsLikelihood(Likelihood):
    r"""Class implementing a Gaussian likelihood model with lower censorship. This likelihood function is introduced in Section II.C of :cite:t:`paludEfficientSamplingNon2023`.

    Note
    ----
    This likelihood is a parametric approximation of the true likelihood model.
    The associated parameter, denoted :math:`a_\ell` in the article (as there is one such parameter per observable :math:`\ell `), should be adjusted before any inversion.
    To adjust this parameter, see the ``beetroots.approx_optim`` subpackage.
    """

    __slots__ = (
        "N",
        "L",
        "D",
        "y",
        "forward_map",
        "log_y",
        "sigma_a",
        "sigma_m",
        "omega",
        "log_fm1",
        "log_fp1",
        "P_lambda",
        "grad_P_lambda",
        "hess_diag_P_lambda",
    )

    def __init__(
        self,
        forward_map,
        D: int,
        L: int,
        N: int,
        y: np.ndarray,
        sigma_a: Union[float, np.ndarray],
        sigma_m: Union[float, np.ndarray],
        omega: Union[float, np.ndarray],
        path_transition_params: str,
        list_lines_fit: List[str],
    ) -> None:
        """Constructor of the GaussianLikelihood object.

        Parameters
        ----------
        forward_map : ForwardMap instance
            forward map
        D : int
            number of disinct physical parameters in input space.
        L : int
            number of distinct observed physical parameters.
        N : int
            number of pixels in each physical dimension
        y : np.ndarray of shape (N, L)
            mean of the gaussian distribution
        sigma_a : float or np.ndarray of shape (N, L)
            standard deviation of the Gaussian distribution
        sigma_m : float or np.ndarray of shape (N, L)
            scale parameter of the lognormal distribution
        omega : float or np.ndarray of shape (N, L)
            censorship threshold

        Raises
        ------
        ValueError
            y must have the shape (N, L)
        """
        super().__init__(forward_map, D, L, N, y)

        # ! trigger an error is the mean y contains less than N elements
        if not (y.shape == (N, L)):
            raise ValueError(
                "y must have the shape (N, L) = ({}, {}) elements".format(
                    self.N, self.L
                )
            )
        if isinstance(sigma_a, (float, int)):
            self.sigma_a = sigma_a * np.ones((N, L))
        else:
            assert sigma_a.shape == (N, L)
            self.sigma_a = sigma_a

        if isinstance(sigma_m, (float, int)):
            self.sigma_m = sigma_m * np.ones((N, L))
        else:
            assert sigma_m.shape == (N, L)
            self.sigma_m = sigma_m

        if isinstance(omega, (float, int)):
            self.omega = omega * np.ones((N, L))
        else:
            assert omega.shape == (N, L)
            self.omega = omega

        self.log_y = np.log(y)

        df_transition = pd.read_csv(path_transition_params)
        num_distinct_n = np.unique(df_transition["n"]).size

        if num_distinct_n == 1:
            # if only one n value, use the transition params for all pixels
            df_transition = df_transition.set_index("line")

            transition_center = (
                df_transition.loc[list_lines_fit, "a0_best"].values[None, :]
                * np.ones((self.N, self.L))
                * np.log(10)
            )
            transition_radius = (
                df_transition.loc[list_lines_fit, "a1_best"].values[None, :]
                * np.ones((self.N, self.L))
                * np.log(10)
            )

        else:
            # assert (
            #     num_distinct_n == self.N
            # ), f"the transition file {path_transition_params} should have {self.N} or 1 distinct values of n, and has {num_distinct_n}"

            df_transition = df_transition.set_index(["n", "line"])
            index = pd.MultiIndex.from_product(
                [list(range(self.N)), list_lines_fit],
                names=["n", "line"],
            )
            # df_transition

            transition_center = df_transition.loc[index, "a0_best"].values.reshape(
                (self.N, self.L)
            ) * np.log(10)
            transition_radius = df_transition.loc[index, "a1_best"].values.reshape(
                (self.N, self.L)
            ) * np.log(10)

        # for col in ["ell", "a0_best", "a1_best", "target_best"]:
        #     assert col in list(df_transition.columns)

        # df_transition = df_transition.set_index("ell")
        # df_transition = df_transition.sort_index()
        # transition_center = (
        #     df_transition.loc[:, "a0_best"].values[None, :]
        #     * np.ones((self.N, self.L))
        #     * np.log(10)
        # )
        # transition_radius = (
        #     df_transition.loc[:, "a1_best"].values[None, :]
        #     * np.ones((self.N, self.L))
        #     * np.log(10)
        # )

        self.log_fm1 = transition_center - transition_radius  # (N, L)
        self.log_fp1 = transition_center + transition_radius  # (N, L)

        self.P_lambda = np.poly1d(np.array([-6.0, 15.0, -10.0, 0.0, 0.0, 1.0]))
        self.grad_P_lambda = self.P_lambda.deriv(m=1)
        self.hess_diag_P_lambda = self.P_lambda.deriv(m=2)

    def sample_observation_model(
        self,
        forward_map_evals: dict,
        rng: np.random.Generator = np.random.default_rng(),
    ) -> np.ndarray:
        eps_a = rng.normal(loc=0.0, scale=self.sigma_a)
        eps_m = rng.lognormal(
            mean=-(self.sigma_m**2) / 2,
            sigma=self.sigma_m,
        )
        f_Theta = forward_map_evals["f_Theta"] * 1
        y_rep = np.maximum(self.omega, eps_m * f_Theta + eps_a)
        return y_rep

    def model_mixing_param(
        self,
        forward_map_evals: dict,
        idx: Optional[np.ndarray] = None,
    ) -> np.ndarray:
        r"""computes the weight of the additive model :math:`\lambda_{n, \ell}` (line-wise and pixel-wise). In this model, :math:`\lambda_{n, \ell}` is a functino of the observation :math:`y_{n, \ell}`, and therefore constant during the sampling

        .. math::
            \lambda_{n, \ell} = \frac{\sigma_a^{2b}}{\sigma_a^{2b} + (a \sigma_m y_{n, \ell})^{2b}}

        with :math:`a` a transition location parameter and :math:`b` a transition speed parameter
        """
        if idx is None:
            N_pix = self.N * 1
            log_fm1 = self.log_fm1 * 1
            log_fp1 = self.log_fp1 * 1
            # sigma_a = self.sigma_a * 1
            # sigma_m = self.sigma_m * 1
        else:
            n_pix = idx.size
            k_mtm = forward_map_evals["f_Theta"].shape[0] // n_pix
            N_pix = forward_map_evals["f_Theta"].shape[0]

            log_fm1 = np.zeros((n_pix, k_mtm, self.L))
            log_fp1 = np.zeros((n_pix, k_mtm, self.L))

            for i_pix in range(n_pix):
                log_fm1[i_pix, :, :] = self.log_fm1[idx[i_pix], :][None, :] * np.ones(
                    (k_mtm, self.L)
                )
                log_fp1[i_pix, :, :] = self.log_fp1[idx[i_pix], :][None, :] * np.ones(
                    (k_mtm, self.L)
                )

            log_fm1 = log_fm1.reshape((N_pix, self.L))
            log_fp1 = log_fp1.reshape((N_pix, self.L))

            # N_pix = forward_map_evals["f_Theta"].shape[0]

            # log_fm1 = np.zeros((N_pix, self.L))
            # log_fp1 = np.zeros((N_pix, self.L))
            # for i in range(idx.size):
            #     log_fm1[i, :] = self.log_fm1[idx[i], :] * 1
            #     log_fp1[i, :] = self.log_fp1[idx[i], :] * 1

            # log_fp1 = self.log_fp1[idx, :][None, :] * np.ones((N_pix, self.L))

            # sigma_a = self.sigma_a[idx, :][None, :] * np.ones((N_pix, self.L))
            # sigma_m = self.sigma_m[idx, :][None, :] * np.ones((N_pix, self.L))

        # lambda_ = 1 / (
        #     1
        #     + np.exp(
        #         (2 * self.transition_speed)
        #         * np.log(
        #             (self.transition_loc * sigma_m * forward_map_evals["f_Theta"]) / sigma_a
        #         )
        #     )
        # )
        lambda_ = np.where(
            forward_map_evals["log_f_Theta"] <= log_fm1,
            1,
            np.where(
                forward_map_evals["log_f_Theta"] >= log_fp1,
                0,
                self.P_lambda(
                    (forward_map_evals["log_f_Theta"] - log_fm1) / (log_fp1 - log_fm1)
                ),
            ),
        )
        return lambda_  # (N, L)

    def grad_model_mixing_param(self, forward_map_evals: dict) -> np.ndarray:
        r"""[summary]

        Parameters
        ----------
        lambda_ : np.ndarray of shape (N, L)
            [description]
        f_Theta : np.ndarray of shape (N, L)
            [description]
        grad_f_Theta : np.ndarray of shape (N, D, L)
            [description]

        Returns
        -------
        np.ndarray of shape (N, D, L)
            [description]
        """
        # grad_ = (
        #     (-2 * self.transition_speed)
        #     * (
        #         (
        #             (self.transition_loc * self.sigma_m / self.sigma_a)
        #             ** (2 * self.transition_speed)
        #         )
        #         * (
        #             forward_map_evals["f_Theta"] ** (2 * self.transition_speed - 1)
        #             * lambda_ ** 2
        #         )
        #     )[:, None, :]
        #     * forward_map_evals["grad_f_Theta"]
        # )
        u = (forward_map_evals["log_f_Theta"] - self.log_fm1) / (
            self.log_fp1 - self.log_fm1
        )  # (N, L)
        grad_u = (
            forward_map_evals["grad_log_f_Theta"]
            / (self.log_fp1 - self.log_fm1)[:, None, :]
        )  # (N, D, L)
        # assert u.shape == (self.N, self.L)
        # assert grad_u.shape == (self.N, self.D, self.L), grad_u.shape

        grad_ = np.where(
            (
                (forward_map_evals["log_f_Theta"] <= self.log_fm1)
                | (forward_map_evals["log_f_Theta"] >= self.log_fp1)
            )[:, None, :],
            np.zeros((self.N, self.D, self.L)),
            grad_u * self.grad_P_lambda(u)[:, None, :],
        )
        assert grad_.shape == (self.N, self.D, self.L), grad_.shape

        return grad_  # (N, D, L)

    def hess_diag_model_mixing_param(self, forward_map_evals: dict) -> np.ndarray:
        r"""[summary]

        Parameters
        ----------
        lambda_ : np.ndarray of shape (N, L)
            [description]
        grad_lambda_ : np.ndarray of shape (N, D, L)
            [description]
        f_Theta : np.ndarray of shape (N, L)
            [description]
        grad_f_Theta : np.ndarray of shape (N, D, L)
            [description]
        hess_diag_f_Theta : np.ndarray of shape (N, D, L)
            [description]

        Returns
        -------
        np.ndarray of shape (N, D, L)
            [description]
        """
        # hess_diag = (
        #     (-2 * self.transition_speed)
        #     * (self.transition_loc * self.sigma_m / self.sigma_a)
        #     ** (2 * self.transition_speed)
        # )[:, None, :] * (
        #     (
        #         forward_map_evals["f_Theta"] ** (2 * self.transition_speed - 1)
        #         * lambda_ ** 2
        #     )[:, None, :]
        #     * forward_map_evals["hess_diag_f_Theta"]
        #     + (2 * self.transition_speed - 1)
        #     * (
        #         forward_map_evals["grad_f_Theta"] ** 2
        #         * (
        #             lambda_ ** 2
        #             * forward_map_evals["f_Theta"] ** (2 * self.transition_speed - 2)
        #         )[:, None, :]
        #     )
        #     + 2
        #     * (forward_map_evals["f_Theta"] ** (2 * self.transition_speed - 1) * lambda_)[
        #         :, None, :
        #     ]
        #     * forward_map_evals["grad_f_Theta"]
        #     * grad_lambda_
        # )

        u = (forward_map_evals["log_f_Theta"] - self.log_fm1) / (
            self.log_fp1 - self.log_fm1
        )  # (N, L)
        grad_u = (
            forward_map_evals["grad_log_f_Theta"]
            / (self.log_fp1 - self.log_fm1)[:, None, :]
        )  # (N, D, L)
        hess_diag_u = (
            forward_map_evals["hess_diag_log_f_Theta"]
            / (self.log_fp1 - self.log_fm1)[:, None, :]
        )  # (N, D, L)

        hess_diag = np.where(
            (
                (forward_map_evals["log_f_Theta"] <= self.log_fm1)
                | (forward_map_evals["log_f_Theta"] >= self.log_fp1)
            )[:, None, :],
            np.zeros((self.N, self.D, self.L)),
            (
                grad_u**2 * self.hess_diag_P_lambda(u)[:, None, :]
                + hess_diag_u * self.grad_P_lambda(u)[:, None, :]
            ),
        )
        assert hess_diag.shape == (self.N, self.D, self.L), hess_diag.shape
        return hess_diag  # (N, D, L)

    def _compute_bias_and_std(
        self,
        forward_map_evals: dict,
        idx: Optional[np.ndarray] = None,
    ) -> np.ndarray:
        r"""computes the biases and std of additive and multiplicative approximations

        Parameters
        ----------
        f_Theta : numpy.array of shape (N_pix, L)
            image of the current iterate through the forward model
        sigma_a : np

        Returns
        -------
        current_bias_std : np.ndarray of shape (4, N_pix, L)
            array with (m_a, s_a, m_m, s_m)
        """
        if idx is None:
            N_pix = self.N * 1
            sigma_a = self.sigma_a * 1
            sigma_m = self.sigma_m * 1

        else:
            N_pix = forward_map_evals["f_Theta"].shape[0]

            sigma_a = np.zeros((N_pix, self.L))
            sigma_m = np.zeros((N_pix, self.L))
            for i in range(idx.size):
                sigma_a[i, :] = self.sigma_a[idx[i], :] * 1
                sigma_m[i, :] = self.sigma_m[idx[i], :] * 1

        log_combination = (
            np.log(sigma_a) - forward_map_evals["log_f_Theta"] - (sigma_m**2) / 2
        )

        assert sigma_a.min() > 0, sigma_a.min()
        assert sigma_m.min() > 0, sigma_m.min()
        assert np.sum(np.isnan(forward_map_evals["log_f_Theta"])) == 0, np.sum(
            np.isnan(forward_map_evals["log_f_Theta"])
        )
        assert np.sum(np.isnan(log_combination)) == 0, np.sum(np.isnan(log_combination))

        # * computation of bias and variances
        m_a = (np.exp(sigma_m**2 / 2) - 1) * forward_map_evals["f_Theta"]
        s_a = sigma_a * np.sqrt(
            (np.exp(sigma_m**2) - 1) * np.exp(-2 * log_combination) + 1
        )

        m_m = -0.5 * np.log(1 + np.exp(2 * log_combination))
        s_m = np.sqrt(sigma_m**2 - 2 * m_m)

        assert s_m.min() > 0, f"{s_m.min()}, {(sigma_m ** 2 - 2 * m_m).min()}"

        # gather all in one array
        N_pix = forward_map_evals["f_Theta"].shape[0]
        current_bias_std = np.zeros((4, N_pix, self.L))
        current_bias_std[0] = m_a
        current_bias_std[1] = s_a
        current_bias_std[2] = m_m
        current_bias_std[3] = s_m

        return current_bias_std

    def neglog_pdf(
        self,
        forward_map_evals: dict,
        nll_utils: dict,
        pixelwise: bool = False,
        full: bool = False,
        idx: Optional[np.ndarray] = None,
    ) -> Union[float, np.ndarray]:

        nlpdf = nll_utils["lambda_"] * np.where(
            nll_utils["censored_mask"], nll_utils["nll_ac"], nll_utils["nll_au"]
        ) + (1 - nll_utils["lambda_"]) * np.where(
            nll_utils["censored_mask"], nll_utils["nll_mc"], nll_utils["nll_mu"]
        )  # (N, L)

        nlpdf = np.nan_to_num(nlpdf)
        nlpdf -= np.log(nll_utils["sigma_a"]) + np.log(nll_utils["sigma_m"])
        # nlpdf /= self.N * self.L
        # nll_utils["censored_mask"].size = self.N * self.L if standard eval
        # nll_utils["censored_mask"].size = N_candidates * self.L if candidates
        if full:
            return nlpdf  # (N, L)

        if pixelwise:
            sum_ = np.sum(nlpdf, axis=1)  # (N,)
            assert sum_.size == forward_map_evals["f_Theta"].shape[0]
            return sum_

        return nlpdf.sum()

    def neglog_pdf_ac(
        self,
        forward_map_evals: dict,
        nll_utils: dict,
        omega: np.ndarray,
    ) -> np.ndarray:
        z = (omega - forward_map_evals["f_Theta"] - nll_utils["m_a"]) / nll_utils["s_a"]
        # z = np.nan_to_num(z)

        nll_ac = -log_ndtr(z)
        nll_ac = np.nan_to_num(nll_ac)
        return nll_ac

    def neglog_pdf_au(
        self,
        forward_map_evals: dict,
        nll_utils: dict,
        y: np.ndarray,
    ) -> np.ndarray:
        nll_au = cost_au(
            y=y,
            f_Theta=forward_map_evals["f_Theta"],
            m_a=nll_utils["m_a"],
            s_a=nll_utils["s_a"],
        )
        # add a constant to `nll_au` to ensure that it is positive
        # nll_au -= np.log(nll_utils["sigma_a"])
        # nll_au = np.nan_to_num(nll_au)
        return nll_au

    def neglog_pdf_mc(
        self,
        forward_map_evals: dict,
        nll_utils: dict,
        log_omega: np.ndarray,
    ) -> np.ndarray:
        z = log_omega - forward_map_evals["log_f_Theta"] - nll_utils["m_m"]
        z /= nll_utils["s_m"]
        # z = np.nan_to_num(z)

        nll_mc = -log_ndtr(z)
        nll_mc = np.nan_to_num(nll_mc)
        return nll_mc

    def neglog_pdf_mu(
        self,
        forward_map_evals: dict,
        nll_utils: dict,
        log_y: np.ndarray,
    ) -> np.ndarray:
        nll_mu = cost_mu(
            log_y=log_y,
            log_f_Theta=forward_map_evals["log_f_Theta"],
            m_m=nll_utils["m_m"],
            s_m=nll_utils["s_m"],
        )
        # nll_mu -= np.log(nll_utils["sigma_m"])
        return nll_mu

    def gradient_neglog_pdf(
        self, forward_map_evals: dict, nll_utils: dict
    ) -> np.ndarray:
        """[summary]

        [extended_summary]

        Parameters
        ----------
        x : np.ndarray of shape (N, D)
            [description]
        f_Theta : np.ndarray of shape (N, L), optional
            image of x via forward map, by default None
        grad_f_Theta : np.ndarray of shape (N, D, L), optional
            [description], by default None

        Returns
        -------
        np.ndarray of shape (N, D, L)
            [description]
        """
        grad_ = np.where(
            (self.y == self.omega)[:, None, :],
            # censored
            # grad_nll_ac,
            nll_utils["lambda_"][:, None, :] * nll_utils["grad_nll_ac"]
            + nll_utils["grad_lambda_"] * nll_utils["nll_ac"][:, None, :]
            + (1 - nll_utils["lambda_"])[:, None, :] * nll_utils["grad_nll_mc"]
            - nll_utils["grad_lambda_"] * nll_utils["nll_mc"][:, None, :],
            # uncensored
            nll_utils["lambda_"][:, None, :] * nll_utils["grad_nll_au"]
            + nll_utils["grad_lambda_"] * nll_utils["nll_au"][:, None, :]
            + (1 - nll_utils["lambda_"])[:, None, :] * nll_utils["grad_nll_mu"]
            - nll_utils["grad_lambda_"] * nll_utils["nll_mu"][:, None, :],
        )  # (N, D, L)
        # grad_ = np.where(
        #     np.isfinite(grad_),
        #     grad_,
        #     np.abs(grad_).max(axis=0)[None, :, :],
        # )  # (N, D, L)
        grad_ = np.nan_to_num(grad_)
        return np.sum(grad_, axis=2)  # / (self.N * self.L)

    def gradient_neglog_pdf_ac(
        self, forward_map_evals: dict, nll_utils: dict
    ) -> np.ndarray:
        f_Theta_m_a_omega = forward_map_evals["f_Theta"] + nll_utils["m_a"] - self.omega

        u_1 = utils.norm_pdf_cdf_ratio(-f_Theta_m_a_omega / nll_utils["s_a"])  # (N, L)

        u_2 = (1 / nll_utils["s_a2"])[:, None, :] * (
            (forward_map_evals["grad_f_Theta"] + nll_utils["grad_m_a"])
            * nll_utils["s_a"][:, None, :]
            - f_Theta_m_a_omega[:, None, :] * nll_utils["grad_s_a"]
        )  # (N, D, L)

        grad_ = u_1[:, None, :] * u_2  # (N, D, L)
        # grad_ = np.nan_to_num(grad_)
        # assert np.sum(np.isnan(grad_)) == 0, f"grad_ac : {np.sum(np.isnan(grad_))}"
        return grad_

    def gradient_neglog_pdf_au(
        self, forward_map_evals: dict, nll_utils: dict
    ) -> np.ndarray:
        g_au = gradient_cost_au(
            y=self.y,
            grad_f_Theta=forward_map_evals["grad_f_Theta"],
            f_Theta=forward_map_evals["f_Theta"],
            grad_m_a=nll_utils["grad_m_a"],
            m_a=nll_utils["m_a"],
            grad_s_a2=nll_utils["grad_s_a2"],
            s_a2=nll_utils["s_a2"],
        )
        # g_au = np.nan_to_num(g_au)
        return g_au

    def gradient_neglog_pdf_mc(
        self, forward_map_evals: dict, nll_utils: dict
    ) -> np.ndarray:
        log_f_Theta_m_m_log_omega = (
            forward_map_evals["log_f_Theta"] + nll_utils["m_m"] - np.log(self.omega)
        )

        u_1 = utils.norm_pdf_cdf_ratio(
            -log_f_Theta_m_m_log_omega / nll_utils["s_m"]
        )  # (N, L)
        u_2 = (1 / nll_utils["s_m2"])[:, None, :] * (
            (forward_map_evals["grad_log_f_Theta"] + nll_utils["grad_m_m"])
            * nll_utils["s_m"][:, None, :]
            - log_f_Theta_m_m_log_omega[:, None, :] * nll_utils["grad_s_m"]
        )  # (N, D, L)

        grad_ = u_1[:, None, :] * u_2  # (N, D, L)
        # grad_ = np.nan_to_num(grad_)
        # assert np.sum(np.isnan(grad_)) == 0, f"grad_ac : {np.sum(np.isnan(grad_))}"
        return grad_

    def gradient_neglog_pdf_mu(
        self, forward_map_evals: dict, nll_utils: dict
    ) -> np.ndarray:
        g_mu = gradient_cost_mu(
            log_y=self.log_y,
            grad_log_f_Theta=forward_map_evals["grad_log_f_Theta"],
            log_f_Theta=forward_map_evals["log_f_Theta"],
            grad_m_m=nll_utils["grad_m_m"],
            m_m=nll_utils["m_m"],
            grad_s_m2=nll_utils["grad_s_m2"],
            s_m2=nll_utils["s_m2"],
        )
        # g_mu = np.nan_to_num(g_mu)
        return g_mu

    def hess_diag_neglog_pdf(
        self, forward_map_evals: dict, nll_utils: dict
    ) -> np.ndarray:
        r"""[summary]

        [extended_summary]

        Parameters
        ----------
        x : np.ndarray of shape (N, D)
            [description]
        f_Theta : np.ndarray of shape (N, L), optional
            [description], by default None
        grad_f_Theta : np.ndarray of shape (N, D, L), optional
            [description], by default None
        hess_diag_f_Theta : np.ndarray of shape (N, D, L), optional
            [description], by default None

        Returns
        -------
        np.ndarray of shape (N, D, L)
            [description]
        """
        hess_diag = np.where(
            (self.y == self.omega)[:, None, :],
            # censored
            # hess_diag_ac,
            (
                nll_utils["lambda_"][:, None, :] * nll_utils["hess_diag_ac"]
                + nll_utils["hess_diag_lambda_"] * nll_utils["nll_ac"][:, None, :]
                + 2 * nll_utils["grad_lambda_"] * nll_utils["grad_nll_ac"]
                #
                + (1 - nll_utils["lambda_"])[:, None, :] * nll_utils["hess_diag_mc"]
                - nll_utils["hess_diag_lambda_"] * nll_utils["nll_mc"][:, None, :]
                - 2 * nll_utils["grad_lambda_"] * nll_utils["grad_nll_mc"]
            ),
            # uncensored
            (
                nll_utils["lambda_"][:, None, :] * nll_utils["hess_diag_au"]
                + nll_utils["hess_diag_lambda_"] * nll_utils["nll_au"][:, None, :]
                + 2 * nll_utils["grad_lambda_"] * nll_utils["grad_nll_au"]
                #
                + (1 - nll_utils["lambda_"])[:, None, :] * nll_utils["hess_diag_mu"]
                - nll_utils["hess_diag_lambda_"] * nll_utils["nll_mu"][:, None, :]
                - 2 * nll_utils["grad_lambda_"] * nll_utils["grad_nll_mu"]
            ),
        )  # (N, D, L)
        # hess_diag = np.where(
        #     np.isfinite(hess_diag),
        #     hess_diag,
        #     np.abs(hess_diag).max(axis=0)[None, :, :],
        # )  # (N, D, L)
        hess_diag = np.nan_to_num(hess_diag)
        return np.sum(hess_diag, axis=2)  # / (self.N * self.L)

    def hess_diag_neglog_pdf_ac(
        self, forward_map_evals: dict, nll_utils: dict
    ) -> np.ndarray:
        f_Theta_m_a_omega = forward_map_evals["f_Theta"] + nll_utils["m_a"] - self.omega
        grad_f_Theta_grad_m_a = (
            forward_map_evals["grad_f_Theta"] + nll_utils["grad_m_a"]
        )

        u_1 = utils.norm_pdf_cdf_ratio(-f_Theta_m_a_omega / nll_utils["s_a"])  # (N, L)
        assert u_1.shape == (self.N, self.L)

        u_2 = (1 / nll_utils["s_a2"])[:, None, :] * (
            grad_f_Theta_grad_m_a * nll_utils["s_a"][:, None, :]
            - f_Theta_m_a_omega[:, None, :] * nll_utils["grad_s_a"]
        )  # (N, D, L)
        assert u_2.shape == (self.N, self.D, self.L)

        grad_u_1 = (
            u_2 * (u_1 * (-f_Theta_m_a_omega / nll_utils["s_a"] + u_1))[:, None, :]
        )  # (N, D, L)
        assert grad_u_1.shape == (self.N, self.D, self.L)

        grad_u_2 = (1 / nll_utils["s_a2"] ** 2)[:, None, :] * (
            nll_utils["s_a2"][:, None, :]
            * (
                nll_utils["s_a"][:, None, :]
                * (forward_map_evals["hess_diag_f_Theta"] + nll_utils["hess_diag_m_a"])
                - f_Theta_m_a_omega[:, None, :] * nll_utils["hess_diag_s_a"]
            )
            - nll_utils["grad_s_a2"]
            * (
                nll_utils["s_a"][:, None, :] * grad_f_Theta_grad_m_a
                - f_Theta_m_a_omega[:, None, :] * nll_utils["grad_s_a"]
            )
        )  # (N, D, L)
        assert grad_u_2.shape == (self.N, self.D, self.L)

        hess_diag = grad_u_1 * u_2 + u_1[:, None, :] * grad_u_2
        # hess_diag = np.nan_to_num(hess_diag)
        return hess_diag

    def hess_diag_neglog_pdf_au(
        self, forward_map_evals: dict, nll_utils: dict
    ) -> np.ndarray:
        f_Theta_m_a_y = (
            forward_map_evals["f_Theta"] + nll_utils["m_a"] - self.y
        )  # (N, L)
        grad_f_Theta_grad_m_a = (
            forward_map_evals["grad_f_Theta"] + nll_utils["grad_m_a"]
        )  # (N, D, L)

        u_1 = f_Theta_m_a_y / nll_utils["s_a2"] ** 2  # (N, L)

        grad_u_1 = (1 / nll_utils["s_a2"] ** 4)[:, None, :] * (
            grad_f_Theta_grad_m_a * (nll_utils["s_a2"] ** 2)[:, None, :]
            - 2
            * (f_Theta_m_a_y * nll_utils["s_a2"])[:, None, :]
            * nll_utils["grad_s_a2"]
        )  # (N, D, L)

        u_2 = (
            2 * nll_utils["s_a2"][:, None, :] * grad_f_Theta_grad_m_a
            - f_Theta_m_a_y[:, None, :] * nll_utils["grad_s_a2"]
        )  # (N, D, L)

        grad_u_2 = (
            2
            * nll_utils["s_a2"][:, None, :]
            * (forward_map_evals["hess_diag_f_Theta"] + nll_utils["hess_diag_m_a"])
            #
            + nll_utils["grad_s_a2"] * grad_f_Theta_grad_m_a
            #
            - f_Theta_m_a_y[:, None, :] * nll_utils["hess_diag_s_a2"]
        )  # (N, D, L)

        hess_diag = (
            0.5
            * (
                nll_utils["hess_diag_s_a2"] * nll_utils["s_a2"][:, None, :]
                - nll_utils["grad_s_a2"] ** 2
            )
            / (nll_utils["s_a2"] ** 2)[:, None, :]
        )  # (N, D, L)

        hess_diag += 0.5 * (
            grad_u_1 * u_2 + u_1[:, None, :] * grad_u_2
        )  # hess_diag = np.nan_to_num(hess_diag)
        return hess_diag

    def hess_diag_neglog_pdf_mc(
        self, forward_map_evals: dict, nll_utils: dict
    ) -> np.ndarray:
        log_f_Theta_m_m_log_omega = (
            forward_map_evals["log_f_Theta"] + nll_utils["m_m"] - self.omega
        )
        grad_log_f_Theta_grad_m_m = (
            forward_map_evals["grad_log_f_Theta"] + nll_utils["grad_m_m"]
        )

        u_1 = utils.norm_pdf_cdf_ratio(
            -log_f_Theta_m_m_log_omega / nll_utils["s_m"]
        )  # (N, L)
        assert u_1.shape == (self.N, self.L)

        u_2 = (1 / nll_utils["s_m2"])[:, None, :] * (
            grad_log_f_Theta_grad_m_m * nll_utils["s_m"][:, None, :]
            - log_f_Theta_m_m_log_omega[:, None, :] * nll_utils["grad_s_m"]
        )  # (N, D, L)
        assert u_2.shape == (self.N, self.D, self.L)

        grad_u_1 = (
            u_2
            * (u_1 * (-log_f_Theta_m_m_log_omega / nll_utils["s_m"] + u_1))[:, None, :]
        )  # (N, D, L)
        assert grad_u_1.shape == (self.N, self.D, self.L)

        grad_u_2 = (1 / nll_utils["s_m2"] ** 2)[:, None, :] * (
            nll_utils["s_m2"][:, None, :]
            * (
                nll_utils["s_m"][:, None, :]
                * (forward_map_evals["hess_diag_f_Theta"] + nll_utils["hess_diag_m_m"])
                - log_f_Theta_m_m_log_omega[:, None, :] * nll_utils["hess_diag_s_m"]
            )
            - nll_utils["grad_s_m2"]
            * (
                nll_utils["s_m"][:, None, :] * grad_log_f_Theta_grad_m_m
                - log_f_Theta_m_m_log_omega[:, None, :] * nll_utils["grad_s_m"]
            )
        )  # (N, D, L)
        assert grad_u_2.shape == (self.N, self.D, self.L)

        hess_diag = grad_u_1 * u_2 + u_1[:, None, :] * grad_u_2
        # hess_diag = np.nan_to_num(hess_diag)
        return hess_diag

    def hess_diag_neglog_pdf_mu(
        self, forward_map_evals: dict, nll_utils: dict
    ) -> np.ndarray:
        log_f_Theta_m_m_log_y = (
            forward_map_evals["log_f_Theta"] + nll_utils["m_m"] - self.log_y
        )  # (N, L)
        grad_log_f_Theta_grad_m_m = (
            forward_map_evals["grad_log_f_Theta"] + nll_utils["grad_m_m"]
        )  # (N, D, L)

        u_1 = log_f_Theta_m_m_log_y / nll_utils["s_m2"] ** 2  # (N, L)

        grad_u_1 = (1 / nll_utils["s_m2"] ** 4)[:, None, :] * (
            grad_log_f_Theta_grad_m_m * (nll_utils["s_m2"] ** 2)[:, None, :]
            - 2
            * (log_f_Theta_m_m_log_y * nll_utils["s_m2"])[:, None, :]
            * nll_utils["grad_s_m2"]
        )  # (N, D, L)

        u_2 = (
            2 * nll_utils["s_m2"][:, None, :] * grad_log_f_Theta_grad_m_m
            - log_f_Theta_m_m_log_y[:, None, :] * nll_utils["grad_s_m2"]
        )  # (N, D, L)

        grad_u_2 = 2 * (
            nll_utils["grad_s_m2"] * grad_log_f_Theta_grad_m_m
            + nll_utils["s_m2"][:, None, :]
            * (forward_map_evals["hess_diag_log_f_Theta"] + nll_utils["hess_diag_m_m"])
        ) - (
            nll_utils["grad_s_m2"] * grad_log_f_Theta_grad_m_m
            + log_f_Theta_m_m_log_y[:, None, :] * nll_utils["hess_diag_s_m2"]
        )  # (N, D, L)

        hess_diag = (
            0.5
            * (
                nll_utils["hess_diag_s_m2"] * nll_utils["s_m2"][:, None, :]
                - nll_utils["grad_s_m2"] ** 2
            )
            / (nll_utils["s_m2"] ** 2)[:, None, :]
        )  # (N, D, L)

        hess_diag += 0.5 * (
            grad_u_1 * u_2 + u_1[:, None, :] * grad_u_2
        )  # hess_diag = np.nan_to_num(hess_diag)
        return hess_diag

    def evaluate_all_nll_utils(
        self,
        forward_map_evals: dict,
        idx: Optional[np.ndarray] = None,
        compute_derivatives: bool = True,
        compute_derivatives_2nd_order: bool = True,
    ) -> dict:
        assert (
            np.sum(np.isnan(forward_map_evals["log_f_Theta"])) == 0
        ), f"before entering bias and std : {np.sum(np.isnan(forward_map_evals['log_f_Theta']))}"

        nll_utils = {}
        # * bias and variance
        if idx is None:
            N_pix = self.N * 1
            sigma_a = self.sigma_a * 1
            sigma_m = self.sigma_m * 1
            y = self.y * 1
            log_y = self.log_y * 1
            omega = self.omega * 1
        else:
            n_pix = idx.size
            k_mtm = forward_map_evals["f_Theta"].shape[0] // n_pix
            N_pix = forward_map_evals["f_Theta"].shape[0]
            assert n_pix * k_mtm == N_pix

            sigma_a = np.zeros((n_pix, k_mtm, self.L))
            sigma_m = np.zeros((n_pix, k_mtm, self.L))
            y = np.zeros((n_pix, k_mtm, self.L))
            omega = np.zeros((n_pix, k_mtm, self.L))

            for i_pix in range(n_pix):
                sigma_a[i_pix, :, :] = self.sigma_a[idx[i_pix], :][None, :] * np.ones(
                    (k_mtm, self.L)
                )
                sigma_m[i_pix, :, :] = self.sigma_m[idx[i_pix], :][None, :] * np.ones(
                    (k_mtm, self.L)
                )
                y[i_pix, :, :] = self.y[idx[i_pix], :][None, :] * np.ones(
                    (k_mtm, self.L)
                )
                omega[i_pix, :, :] = self.omega[idx[i_pix], :][None, :] * np.ones(
                    (k_mtm, self.L)
                )

            sigma_a = (
                sigma_a.transpose((2, 0, 1)).reshape((self.L, N_pix)).T
            )  # .reshape((N_pix, self.L))
            sigma_m = (
                sigma_m.transpose((2, 0, 1)).reshape((self.L, N_pix)).T
            )  # .reshape((N_pix, self.L))
            y = (
                y.transpose((2, 0, 1)).reshape((self.L, N_pix)).T
            )  # .reshape((N_pix, self.L))
            log_y = np.log(y)
            omega = (
                omega.transpose((2, 0, 1)).reshape((self.L, N_pix)).T
            )  # .reshape((N_pix, self.L))            omega

        # * -----
        nll_utils["censored_mask"] = (y <= omega) * 1
        nll_utils["sigma_a"] = sigma_a * 1
        nll_utils["sigma_m"] = sigma_m * 1

        sigma_m2 = sigma_m**2
        log_combination = (
            np.log(sigma_a) - forward_map_evals["log_f_Theta"] - sigma_m2 / 2
        )
        exp_m2_log_combin = np.exp(-2 * log_combination)
        exp_sigma_m_squared = np.exp(sigma_m2)
        # exp_sigma_m_squared_div2_m1 = np.exp(sigma_m2 / 2) - 1  # (N, L)

        # * computation of bias and variances
        nll_utils["m_a"] = np.zeros_like(y)
        nll_utils["s_a"] = sigma_a * np.sqrt(
            (exp_sigma_m_squared - 1)
            * np.exp(2 * (forward_map_evals["log_f_Theta"] - np.log(sigma_a)))
            + 1
        )
        nll_utils["s_a2"] = nll_utils["s_a"] ** 2

        nll_utils["m_m"] = -0.5 * (sigma_m2 + np.log(1 + 1 / exp_m2_log_combin))

        nll_utils["s_m2"] = -2 * nll_utils["m_m"]
        nll_utils["s_m"] = np.sqrt(nll_utils["s_m2"])

        assert (
            nll_utils["s_m"].min() > 0
        ), f"{nll_utils['s_m'].min()}, {(sigma_m2 - 2 * nll_utils['m_m']).min()}"

        if compute_derivatives:
            nll_utils["grad_m_a"] = np.zeros((N_pix, self.D, self.L))

            if compute_derivatives_2nd_order:
                nll_utils["hess_diag_m_a"] = np.zeros((N_pix, self.D, self.L))

            nll_utils["grad_s_a2"] = (
                2
                * (forward_map_evals["f_Theta"] * (exp_sigma_m_squared - 1))[:, None, :]
                * forward_map_evals["grad_f_Theta"]
            )  # (N, D, L)

            if compute_derivatives_2nd_order:
                nll_utils["hess_diag_s_a2"] = (
                    2
                    * (exp_sigma_m_squared - 1)[:, None, :]
                    * (
                        forward_map_evals["f_Theta"][:, None, :]
                        * forward_map_evals["hess_diag_f_Theta"]
                        + forward_map_evals["grad_f_Theta"] ** 2
                    )
                )  # (N, D, L)

            nll_utils["grad_s_a"] = (
                1 / (2 * nll_utils["s_a"])[:, None, :]
            ) * nll_utils["grad_s_a2"]

            if compute_derivatives_2nd_order:
                nll_utils["hess_diag_s_a"] = (1 / (2 * nll_utils["s_a2"]))[
                    :, None, :
                ] * (
                    nll_utils["hess_diag_s_a2"] * nll_utils["s_a"][:, None, :]
                    - nll_utils["grad_s_a2"] * nll_utils["grad_s_a"]
                )  # (N, D, L)

            nll_utils["grad_m_m"] = (
                1 / (1 + exp_m2_log_combin) / forward_map_evals["f_Theta"]
            )[:, None, :] * forward_map_evals[
                "grad_f_Theta"
            ]  # (N, D, L)

            if compute_derivatives_2nd_order:
                nll_utils["hess_diag_m_m"] = (
                    1 / (forward_map_evals["f_Theta"] * (1 + exp_m2_log_combin)) ** 2
                )[:, None, :] * (
                    forward_map_evals["hess_diag_f_Theta"]
                    * (forward_map_evals["f_Theta"] * (1 + exp_m2_log_combin))[
                        :, None, :
                    ]
                    - forward_map_evals["grad_f_Theta"] ** 2
                    * (1 + 3 * exp_m2_log_combin)[:, None, :]
                )  # (N, D, L)

            nll_utils["grad_s_m2"] = -2 * nll_utils["grad_m_m"]  # (N, D, L)

            if compute_derivatives_2nd_order:
                nll_utils["hess_diag_s_m2"] = (
                    -2 * nll_utils["hess_diag_m_m"]
                )  # (N, D, L)

            nll_utils["grad_s_m"] = (1 / (2 * nll_utils["s_m"]))[
                :, None, :
            ] * nll_utils[
                "grad_s_m2"
            ]  # (N, D, L)

            if compute_derivatives_2nd_order:
                nll_utils["hess_diag_s_m"] = (1 / (2 * nll_utils["s_m2"]))[
                    :, None, :
                ] * (
                    nll_utils["hess_diag_s_m2"] * nll_utils["s_m"][:, None, :]
                    - nll_utils["grad_s_m"] * nll_utils["grad_s_m2"]
                )  # (N, D, L)

            # assert np.sum(np.isnan(forward_map_evals["f_Theta"])) == 0.0
            # assert (
            #     forward_map_evals["f_Theta"].min() > 0.0
            # ), f"{forward_map_evals['f_Theta'].min()}, {forward_map_evals['log_f_Theta'].min()}"
            # assert np.sum(np.isnan(forward_map_evals["grad_f_Theta"])) == 0.0
            # assert np.sum(np.isnan(forward_map_evals["hess_diag_f_Theta"])) == 0.0

            # assert np.sum(np.isnan(forward_map_evals["log_f_Theta"])) == 0.0
            # assert np.sum(np.isnan(forward_map_evals["grad_log_f_Theta"])) == 0.0
            # assert np.sum(np.isnan(forward_map_evals["hess_diag_log_f_Theta"])) == 0.0

            # assert np.sum(np.isnan(m_a)) == 0.0
            # assert np.sum(np.isnan(grad_m_a)) == 0.0
            # assert np.sum(np.isnan(hess_diag_m_a)) == 0.0
            # assert m_a.shape == (self.N, self.L)
            # assert grad_m_a.shape == (self.N, self.D, self.L)
            # assert hess_diag_m_a.shape == (self.N, self.D, self.L)

            # assert np.sum(np.isnan(s_a2)) == 0.0
            # assert np.all(s_a2 > 0)
            # assert np.sum(np.isnan(grad_s_a2)) == 0.0
            # assert np.sum(np.isnan(hess_diag_s_a2)) == 0.0
            # assert s_a2.shape == (self.N, self.L)
            # assert grad_s_a2.shape == (self.N, self.D, self.L)
            # assert hess_diag_s_a2.shape == (self.N, self.D, self.L)

            # assert np.sum(np.isnan(s_a)) == 0.0
            # assert np.all(s_a > 0)
            # assert np.sum(np.isnan(grad_s_a)) == 0.0
            # assert np.sum(np.isnan(hess_diag_s_a)) == 0.0
            # assert s_a.shape == (self.N, self.L)
            # assert grad_s_a.shape == (self.N, self.D, self.L)
            # assert hess_diag_s_a.shape == (self.N, self.D, self.L)

            # assert np.sum(np.isnan(m_m)) == 0.0
            # assert np.sum(np.isnan(grad_m_m)) == 0.0
            # assert np.sum(np.isnan(hess_diag_m_m)) == 0.0
            # assert m_m.shape == (self.N, self.L)
            # assert grad_m_m.shape == (self.N, self.D, self.L)
            # assert hess_diag_m_m.shape == (self.N, self.D, self.L)

            # assert np.sum(np.isnan(s_m2)) == 0.0
            # assert np.sum(np.isnan(grad_s_m2)) == 0.0
            # assert np.sum(np.isnan(hess_diag_s_m2)) == 0.0
            # assert s_m2.shape == (self.N, self.L)
            # assert grad_s_m2.shape == (self.N, self.D, self.L)
            # assert hess_diag_s_m2.shape == (self.N, self.D, self.L)

            # assert np.sum(np.isnan(s_m)) == 0.0
            # assert np.sum(np.isnan(grad_s_m)) == 0.0
            # assert np.sum(np.isnan(hess_diag_s_m)) == 0.0
            # assert s_m.shape == (self.N, self.L)
            # assert grad_s_m.shape == (self.N, self.D, self.L)
            # assert hess_diag_s_m.shape == (self.N, self.D, self.L)

        # * mixture weight
        nll_utils["lambda_"] = self.model_mixing_param(forward_map_evals, idx)

        if compute_derivatives:
            nll_utils["grad_lambda_"] = self.grad_model_mixing_param(forward_map_evals)
            if compute_derivatives_2nd_order:
                nll_utils["hess_diag_lambda_"] = self.hess_diag_model_mixing_param(
                    forward_map_evals
                )

        nll_utils["nll_au"] = self.neglog_pdf_au(forward_map_evals, nll_utils, y)
        nll_utils["nll_ac"] = self.neglog_pdf_ac(forward_map_evals, nll_utils, y)
        nll_utils["nll_mu"] = self.neglog_pdf_mu(forward_map_evals, nll_utils, log_y)
        nll_utils["nll_mc"] = self.neglog_pdf_mc(
            forward_map_evals, nll_utils, np.log(omega)
        )

        if compute_derivatives:
            nll_utils["grad_nll_au"] = self.gradient_neglog_pdf_au(
                forward_map_evals, nll_utils
            )
            nll_utils["grad_nll_mu"] = self.gradient_neglog_pdf_mu(
                forward_map_evals, nll_utils
            )
            nll_utils["grad_nll_ac"] = self.gradient_neglog_pdf_ac(
                forward_map_evals, nll_utils
            )
            nll_utils["grad_nll_mc"] = self.gradient_neglog_pdf_mc(
                forward_map_evals, nll_utils
            )

            if compute_derivatives_2nd_order:
                nll_utils["hess_diag_ac"] = self.hess_diag_neglog_pdf_ac(
                    forward_map_evals, nll_utils
                )
                nll_utils["hess_diag_au"] = self.hess_diag_neglog_pdf_au(
                    forward_map_evals, nll_utils
                )
                nll_utils["hess_diag_mc"] = self.hess_diag_neglog_pdf_mc(
                    forward_map_evals, nll_utils
                )
                nll_utils["hess_diag_mu"] = self.hess_diag_neglog_pdf_mu(
                    forward_map_evals, nll_utils
                )

        return nll_utils



"""Implementation of log-normal likelihood
"""

from typing import Optional, Union

import numpy as np

from beetroots.modelling.likelihoods.abstract_likelihood import Likelihood


class LogNormalLikelihood(Likelihood):
    """Class implementing a log-normal likelihood model."""

    __slots__ = (
        "forward_map",
        "D",
        "L",
        "N",
        "y",
        "logy",
        "sigma",
    )

    def __init__(
        self,
        forward_map,
        D: int,
        L: int,
        N: int,
        y: np.ndarray,
        sigma: Union[float, np.ndarray],
    ) -> None:
        """Constructor of the LogNormalLikelihood object.

        Parameters
        ----------
        forward_map : ForwardMap instance
            forward map, involved in the mean of the distribution.
        D : int
            number of disinct physical parameters in input space.
        L : int
            number of distinct observed physical parameters.
        N : int
            number of pixels in each physical dimension
        y : np.ndarray of shape (N, L)
            parameter of the log-normal distribution
        sigma : float or np.ndarray of shape (N, L)
            variance of the log-normal distribution

        Raises
        ------
        ValueError
            y must have the shape (N, L)

        Note
        ----
        * Derivatives and Hessians are taken with respect of the mean of the distribution.
        * y provided in log space already? (saving computations)

        """

        # TODO: add method to update y? (instead of having to reinstantiate the full object any time y is updated?)

        super().__init__(forward_map, D, L, N, y)
        self.logy = np.log(self.y)

        # ! trigger an error is the mean y contains less than N elements
        if not y.shape == (N, L):
            raise ValueError(
                "y must have the shape (N, L) = ({}, {}) elements".format(
                    self.N, self.L
                )
            )
        if isinstance(sigma, (float, int)):
            self.sigma = sigma * np.ones(
                (N, L)
            )  # ! P.-A.: not sure this is actually needed... (unless broadcast is not enough here)
        else:
            assert sigma.shape == (N, L)
            self.sigma = sigma

    def _update_observations(self, y):
        """Update the parameters on which the distribution is defined (if
        updated within the solver).

        Parameters
        ----------
        y : np.ndarray of shape (N, L)
            parameter of the log-normal distribution

        Raises
        ------
        ValueError
            y must have the shape (N, L)
        """
        # ! trigger an error is the mean y contains less than N elements
        if not y.shape == (self.N, self.L):
            raise ValueError(
                "y must have the shape (N, L) = ({}, {}) elements".format(
                    self.N, self.L
                )
            )
        self.y = y
        self.logy = np.log(self.y)

    def neglog_pdf(
        self,
        forward_map_evals: dict,
        nll_utils: dict,
        pixelwise: bool = False,
        full: bool = False,
        idx: Optional[np.ndarray] = None,
    ) -> Union[float, np.ndarray]:
        # TODO: there are a few steps to be clarified in there
        # TODO: (what is the point of the reformatting step in here)
        # a priori, sam echange expected
        if idx is None:
            N_pix = self.N * 1
            logy = self.logy * 1
            sigma = self.sigma * 1
        else:
            n_pix = idx.size
            k_mtm = forward_map_evals["f_Theta"].shape[0] // n_pix
            N_pix = forward_map_evals["f_Theta"].shape[0]

            logy = np.zeros((n_pix, k_mtm, self.L))
            sigma = np.zeros((n_pix, k_mtm, self.L))

            for i_pix in range(n_pix):
                logy[i_pix, :, :] = self.logy[idx[i_pix], :][None, :] * np.ones(
                    (k_mtm, self.L)
                )
                sigma[i_pix, :, :] = self.sigma[idx[i_pix], :][None, :] * np.ones(
                    (k_mtm, self.L)
                )

            logy = logy.reshape((N_pix, self.L))
            sigma = sigma.reshape((N_pix, self.L))

        nlpdf = logy + (logy - forward_map_evals["log_f_Theta"]) ** 2 / (
            2 * sigma**2
        )  # (N_pix, L)

        if full:
            return nlpdf  # (N_pix, L)

        if pixelwise:
            return np.sum(nlpdf, axis=1)  # (N_pix,)

        return np.sum(nlpdf)  # float

    def gradient_neglog_pdf(
        self, forward_map_evals: dict, nll_utils: dict
    ) -> np.ndarray:
        grad_ = (
            forward_map_evals["grad_log_f_Theta"]
            * ((forward_map_evals["log_f_Theta"] - self.logy) / self.sigma**2)[
                :, None, :
            ]
        )  # (N, D, L)

        # ! issue: do not sum over L if L = D (i.e. identity forward_map)
        if not self.D == self.L:
            grad_ = np.sum(grad_, axis=2)  # (N, D)

        return grad_

    def hess_diag_neglog_pdf(
        self, forward_map_evals: dict, nll_utils: dict
    ) -> np.ndarray:
        r"""Hessian w.r.t to the parameter of the log-normal distribution.

        Parameters
        ----------
        x : np.ndarray of shape (N, D)
            [description]
        f_Theta : np.ndarray of shape (N, L), optional
            [description], by default None
        grad_f_Theta : np.ndarray of shape (N, D, L), optional
            [description], by default None
        hess_diag_f_Theta : np.ndarray of shape (N, D, L), optional
            [description], by default None

        Returns
        -------
        np.ndarray of shape (N, D)
            [description]
        """
        hess_diag = (1 / self.sigma**2)[:, None, :] * (
            forward_map_evals["grad_log_f_Theta"] ** 2
            + forward_map_evals["hess_diag_log_f_Theta"]
            * (forward_map_evals["f_Theta"] - self.logy)[:, None, :]
        )

        # ! issue: do not sum over L if L = D (i.e. identity forward_map)
        if not self.D == self.L:
            hess_diag = np.sum(hess_diag, axis=2)  # (N, D)

        return hess_diag

    def evaluate_all_forward_map(
        self,
        Theta: np.ndarray,
        compute_derivatives: bool,
        compute_derivatives_2nd_order: bool = True,
    ) -> dict:
        assert len(Theta.shape) == 2 and Theta.shape[1] == self.D
        forward_map_evals = self.forward_map.compute_all(
            Theta, True, True, compute_derivatives, compute_derivatives_2nd_order
        )
        return forward_map_evals

    def evaluate_all_nll_utils(
        self,
        forward_map_evals: dict,
        idx: Optional[int] = None,
        compute_derivatives: bool = False,
        compute_derivatives_2nd_order: bool = True,
    ) -> dict:
        nll_utils = {}
        return nll_utils



r"""Contains a mock class for the scaler object, that does not transform the arrays it receives
"""
import numpy as np

from beetroots.space_transform.abstract_transform import Scaler


class IdScaler(Scaler):

    __slots__ = ()

    def __init__(self):
        pass

    def from_scaled_to_lin(self, Theta_scaled: np.ndarray) -> np.ndarray:
        return Theta_scaled

    def from_lin_to_scaled(self, Theta_linscale: np.ndarray) -> np.ndarray:
        return Theta_linscale



import os
from typing import Dict

import matplotlib.pyplot as plt
import numpy as np
from matplotlib import colors

from beetroots.inversion.plots import readable_line_names
from beetroots.inversion.plots.map_shaper import MapShaper
from beetroots.inversion.plots.plots_2d import AbstractPlots2D


class Plots2DSetup(AbstractPlots2D):
    r"""utilitary class that draws and saves maps related to the inference setup."""

    __slots__ = (
        "path_img",
        "map_shaper",
        "N",
        "pixels_of_interest_names",
        "pixels_of_interest_coords",
    )

    def __init__(
        self,
        path_img: str,
        map_shaper: MapShaper,
        N: int,
        pixels_of_interest: Dict[int, str] = {},
    ):
        self.path_img = path_img
        r"""str: path to the folder in which the figures are to be saved"""

        self.map_shaper = map_shaper
        r"""MapShaper: defines the transformation from vectors to 2D maps"""

        self.N = N
        r"""int: number of pixels in the maps, i.e., dimension of the observation vectors"""

        self.pixels_of_interest_names = pixels_of_interest
        r"""dict: (coordinate, name) pair of some user-informed pixels to be highlighted. These pixels will be outlines with a black square in figures."""

        coords = map_shaper.from_vector_idx_to_map_coords(
            list(pixels_of_interest.keys())
        )
        self.pixels_of_interest_coords = coords
        r"""list: coordinates of some user-informed pixels to be highlighted. These pixels will be outlines with a black square in figures."""

    def plot_indices_map(self):
        r"""plots and saved a map of the indices. Simplifies the choice of pixels of interest."""
        Theta_idx = np.arange(self.N)
        Theta_idx_shaped = self.map_shaper.from_vector_to_map(Theta_idx)

        fig, ax = plt.subplots(figsize=(8, 6))
        plt.title("indices map")
        plt.imshow(
            Theta_idx_shaped,
            origin="lower",
            cmap="jet",
        )

        # Loop over data dimensions and create text annotations.
        font_size = 100 / Theta_idx_shaped.shape[0]
        for i in range(Theta_idx_shaped.shape[0]):
            for j in range(Theta_idx_shaped.shape[1]):
                if not np.isnan(Theta_idx_shaped[i, j]):
                    pix_idx_val = int(Theta_idx_shaped[i, j])
                    _ = ax.text(
                        j,
                        i,
                        f"{pix_idx_val}",
                        ha="center",
                        va="center",
                        color="w",
                        fontsize=font_size,
                    )

        self._draw_rect_on_pixels_of_interest()

        # plt.xticks([]) # one needs the ticks to identify pixel positions
        # plt.yticks([])
        plt.grid()
        plt.savefig(
            f"{self.path_img}/indices_map.PNG",
            bbox_inches="tight",
            dpi=1200,
        )
        plt.close()

    def plot_sites_map(self, dict_sites_: dict[int, np.ndarray]) -> None:
        """plots the map of sites defined from the spatial prior. The pixels within one site are all sampled in parallel in the MTM-chromatic Gibbs sampling kernel (see ``beetroots.sampler.my_sampler.MySampler``)

        Parameters
        ----------
        dict_sites_ : dict[int, np.ndarray]
            dictionary of sites (see spatial prior modules).
        """
        Theta_sites = np.zeros((self.N,))
        for key, idx in dict_sites_.items():
            Theta_sites[idx] = key * 1
        Theta_sites_shaped = self.map_shaper.from_vector_to_map(Theta_sites)

        fig, ax = plt.subplots(figsize=(8, 6))
        plt.title("Sites map")
        plt.imshow(Theta_sites_shaped, origin="lower")
        self._draw_rect_on_pixels_of_interest()

        plt.xticks([])
        plt.yticks([])
        plt.savefig(
            f"{self.path_img}/sites_map.PNG",
            bbox_inches="tight",
        )
        plt.close()

    def plot_censored_lines_proportion(
        self,
        y: np.ndarray,
        omega: np.ndarray,
        folder_path: str,
    ):
        r"""plots the map of the proportion of censored observables per pixel. Only relevant for likelihood models involving censorship.

        Parameters
        ----------
        y : np.ndarray of shape (N, L)
            observation vector of :math:`N` pixels and :math:`L` lines.
        omega : np.ndarray of shape (N, L)
            vector of censorship limits.
        folder_path : str
            path to the folder where the figure is to be saved
        """
        prop_censor = (y <= omega).mean(1)  # float between [0,1]
        prop_censor_shaped = self.map_shaper.from_vector_to_map(prop_censor)

        plt.figure(figsize=(8, 6))
        plt.title("Proportion of censored lines")
        plt.imshow(prop_censor_shaped, origin="lower", cmap="jet")
        plt.colorbar()
        self._draw_rect_on_pixels_of_interest()

        # plt.tight_layout()
        plt.savefig(
            f"{folder_path}/proportion_censored_lines.PNG",
            bbox_inches="tight",
        )
        plt.close()

    def plot_observations(
        self,
        y: np.ndarray,
        list_lines: list,
        folder_path: str,
    ) -> None:
        """list_lines = self.list_lines_fit + self.list_lines_eval"""
        y_shaped = self.map_shaper.from_vector_to_map(y)
        for ell, name in enumerate(list_lines):
            y_ell_shaped = y_shaped[:, :, ell] * 1
            y_ell_shaped[y_ell_shaped >= 0.9] = np.nan
            y_ell_shaped[y_ell_shaped < 1e-14] = np.nan

            readable_name = readable_line_names.lines_to_latex(name)

            plt.figure(figsize=(8, 6))
            plt.title(f"observation of line {readable_name}")
            plt.imshow(
                y_ell_shaped,
                origin="lower",
                norm=colors.LogNorm(),
                cmap="jet",
            )
            plt.colorbar()
            self._draw_rect_on_pixels_of_interest()

            # plt.tight_layout()
            plt.savefig(
                f"{folder_path}/observation_line_{ell}_{name}.PNG",
                bbox_inches="tight",
            )
            plt.close()

            plt.figure(figsize=(8, 6))
            plt.title(f"observation of line {readable_name}")
            plt.imshow(
                y_ell_shaped,
                origin="lower",
                cmap="jet",
            )
            plt.colorbar()
            self._draw_rect_on_pixels_of_interest()

            # plt.tight_layout()
            plt.savefig(
                f"{folder_path}/observation_line_{ell}_{name}_linscale.PNG",
                bbox_inches="tight",
            )
            plt.close()

    def plot_mask_censored_pixels(
        self,
        y: np.ndarray,
        omega: np.ndarray,
        list_lines: list,
        folder_path: str,
    ) -> None:
        """plots and saves the map of censored pixels for each line

        Parameters
        ----------
        y : np.ndarray of shape (N, L)
            observation vector of :math:`N` pixels and :math:`L` lines.
        omega : np.ndarray of shape (N, L)
            vector of censorship limits.
        list_lines : list
            list of the names of the observed lines
        folder_path : str
            path to the folder where the figure is to be saved
        """
        censored_shaped = self.map_shaper.from_vector_to_map(y <= omega)
        for ell, name in enumerate(list_lines):
            censored_ell_shaped = censored_shaped[:, :, ell] * 1

            readable_name = readable_line_names.lines_to_latex(name)

            plt.figure(figsize=(8, 6))
            plt.title(f"mask of censored pixels for line {readable_name}")
            plt.imshow(censored_ell_shaped, origin="lower", cmap="jet")
            plt.colorbar()
            self._draw_rect_on_pixels_of_interest()

            # plt.tight_layout()
            plt.savefig(
                f"{folder_path}/censored_map_line_{ell}_{name}.PNG",
                bbox_inches="tight",
            )
            plt.close()

    def plot_sigma_a(
        self,
        sigma_a: np.ndarray,
        list_lines: list,
        folder_path: str,
    ) -> None:
        r"""plots and saves the maps of standard deviations for each of the :math:`L` observables.

        Parameters
        ----------
        sigma_a : np.ndarray of shape (N, L)
            vector containing the additive noise standard deviations of each of the :math:`N` pixels and :math:`L` lines.
        list_lines : list
            list of the names of the observed lines
        folder_path : str
            path to the folder where the figures are to be saved
        """
        sigma_a_shaped = self.map_shaper.from_vector_to_map(sigma_a)
        for ell, name in enumerate(list_lines):
            sigma_a_ell_shaped = sigma_a_shaped[:, :, ell] * 1
            sigma_a_ell_shaped[sigma_a_ell_shaped >= 0.9] = np.nan

            readable_name = readable_line_names.lines_to_latex(name)

            plt.figure(figsize=(8, 6))
            plt.title(f"standard deviation for line {readable_name}")
            plt.imshow(
                sigma_a_ell_shaped,
                origin="lower",
                norm=colors.LogNorm(),
                cmap="jet",
            )
            plt.colorbar()
            self._draw_rect_on_pixels_of_interest()

            # plt.tight_layout()
            plt.savefig(
                f"{folder_path}/add_err_std_line_{ell}_{name}.PNG",
                bbox_inches="tight",
            )
            plt.close()

    def plot_snr_add(
        self,
        y: np.ndarray,
        sigma_a: np.ndarray,
        list_lines: list,
        folder_path: str,
    ) -> None:
        """plots and saves the maps of signal-to-noise ratio (SNR) for each of the :math:`L` observables. For one pixel :math:`n` and one line :math:`\ell`, the SNR is defined as :math:`y_{n\ell} / \sigma_{a,n\ell}` with :math:`y_{n\ell}` the observed value and :math:`\sigma_{a,n\ell}` the additive noise standard deviation.

        Parameters
        ----------
        y : np.ndarray of shape (N, L)
            observation vector of :math:`N` pixels and :math:`L` lines.
        sigma_a : np.ndarray of shape (N, L)
            vector containing the additive noise standard deviations of each of the :math:`N` pixels and :math:`L` lines.
        list_lines : list
            list of the names of the observed lines
        folder_path : str
            path to the folder where the figures are to be saved
        """
        snr_a_shaped = self.map_shaper.from_vector_to_map(y / sigma_a)
        for ell, name in enumerate(list_lines):
            snr_a_ell_shaped = snr_a_shaped[:, :, ell] * 1

            readable_name = readable_line_names.lines_to_latex(name)

            plt.figure(figsize=(8, 6))
            plt.title(r"y / $\sigma_a$" + f" of line {readable_name}")
            plt.imshow(
                snr_a_ell_shaped,
                origin="lower",
                norm=colors.LogNorm(vmin=1.0),
                cmap="jet",
            )
            plt.colorbar()
            self._draw_rect_on_pixels_of_interest()

            # plt.tight_layout()
            plt.savefig(
                f"{folder_path}/snr_line_{ell}_{name}.PNG",
                bbox_inches="tight",
            )
            plt.close()

    def plot_all(
        self,
        y: np.ndarray,
        sigma_a: np.ndarray,
        omega: np.ndarray,
        list_lines: list,
        dict_sites_: dict[int, np.ndarray],
    ) -> None:
        r"""runs all the class methods, i.e., plots and saves the maps of the observations :math:`y`, of the additive noise standard deviation :math:`\sigma_a`, of the signal-to-noise ratio :math:`y : \sigma_a`, of censored pixels.
        It also plots and saves the map of proportion of censored observables per pixel, the map of indices and of spatial prior sites.

        Parameters
        ----------
        y : np.ndarray of shape (N, L)
            observation vector of :math:`N` pixels and :math:`L` lines.
        sigma_a : np.ndarray of shape (N, L)
            vector containing the additive noise standard deviations of each of the :math:`N` pixels and :math:`L` lines.
        omega : np.ndarray of shape (N, L)
            vector of censorship limits.
        list_lines : list
            list of the names of the observed lines
        dict_sites_ : dict[int, np.ndarray]
            dictionary of sites (see spatial prior modules).
        """
        folder_path = f"{self.path_img}/observations"
        if not os.path.isdir(folder_path):
            os.mkdir(folder_path)

        self.plot_observations(y, list_lines, folder_path)
        self.plot_sigma_a(sigma_a, list_lines, folder_path)
        self.plot_snr_add(y, sigma_a, list_lines, folder_path)
        self.plot_mask_censored_pixels(y, omega, list_lines, folder_path)
        self.plot_censored_lines_proportion(y, omega, folder_path)

        self.plot_indices_map()
        self.plot_sites_map(dict_sites_)



from abc import abstractmethod

import numpy as np

from beetroots.modelling.forward_maps.abstract_base import ForwardMap


class ExpForwardMap(ForwardMap):
    r"""Forward model such that for every pixel :math:`n \in [1, N]`

    .. math::

        f : \theta_n \in \mathbb{R}^D \mapsto f(\theta_n) \in \mathbb{R}^L

    where :math:`f(\theta_n)` can be written with a left composition with the exponential, i.e., :math:`f(\theta_n) = \exp \circ (\ln f) (\theta_n)`
    """

    @abstractmethod
    def evaluate_log(self, Theta: np.ndarray) -> np.ndarray:
        r"""evaluates the log-forward map for a set of input vectors :math:`(\theta_n))_{n=1}^N`

        Parameters
        ----------
        Theta : np.ndarray of shape (N, D)
            array of points in the input space :math:`\Theta = (\theta_n)_{n=1}^N` with :math:`\theta_n \in \mathbb{R}^D`

        Returns
        -------
        np.ndarray of shape (N, L)
            array of the log-forward map images in the observation space :math:`(\ln f(\theta_n))_{n=1}^N` with :math:`\ln f(\theta_n) \in \mathbb{R}^L`
        """
        pass

    @abstractmethod
    def gradient_log(self, Theta: np.ndarray) -> np.ndarray:
        r"""returns the gradient of the log-forward map for a set of input vectors :math:`(\theta_n))_{n=1}^N`

        Parameters
        ----------
        Theta : np.ndarray of shape (N, D)
            array of points in the input space :math:`\Theta = (\theta_n)_{n=1}^N` with :math:`\theta_n \in \mathbb{R}^D`

        Returns
        -------
        np.ndarray of shape (N, D, L)
            array of gradients :math:`(\nabla \ln f(\theta_n))_{n=1}^N`, with :math:`\nabla \ln  f(\theta_n) \in \mathbb{R}^{D \times L}`
        """
        pass

    @abstractmethod
    def hess_diag_log(self, Theta: np.ndarray) -> np.ndarray:
        r"""returns the diagonal of the hessian of the log-forward map for a set of input vectors :math:`(\theta_n))_{n=1}^N`

        Parameters
        ----------
        Theta : np.ndarray of shape (N, D)
            array of points in the input space :math:`\Theta = (\theta_n)_{n=1}^N` with :math:`\theta_n \in \mathbb{R}^D`

        Returns
        -------
        np.ndarray of shape (N, D, L)
            array of diagonal of the hessians :math:`(\text{diag} \nabla^2 \ln f(\theta_n))_{n=1}^N`, with :math:`\text{diag} \nabla^2 \ln f(\theta_n) \in \mathbb{R}^{D \times L}`
        """
        pass



import copy
import multiprocessing as mp
import os
import time
from concurrent.futures import ProcessPoolExecutor
from typing import Optional, Union

import numpy as np
import pandas as pd

from beetroots.inversion.results.results_optim_map import ResultsExtractorOptimMAP
from beetroots.inversion.results.results_optim_mle import ResultsExtractorOptimMLE
from beetroots.inversion.run.abstract_run import Run
from beetroots.modelling.posterior import Posterior
from beetroots.sampler.abstract_sampler import Sampler
from beetroots.sampler.saver.abstract_saver import Saver
from beetroots.space_transform.abstract_transform import Scaler


class RunMCMC(Run):
    r"""class that runs inversions using a sampling approach"""

    __slots__ = ("path_data_csv_out", "max_workers")

    def __init__(self, path_data_csv_out: str, max_workers: int):
        r"""

        Parameters
        ----------
        path_data_csv_out : str
            path to the folder where results are to be saved
        max_workers : int
            max number of workers that can be used to run the inversion
        """
        self.path_data_csv_out = path_data_csv_out
        r"""str: path to the folder where results are to be saved"""

        self.max_workers = max_workers
        r"""int: max number of workers that can be used to run the inversion"""

    def prepare_run(
        self,
        dict_posteriors: dict[str, Posterior],
        path_raw: str,
        N_runs: int,
        scaler: Scaler,
        start_from: Optional[str],
        path_csv_mle: Optional[str],
        path_csv_map: Optional[str],
    ) -> Optional[np.ndarray]:
        r"""prepares the run in two ways :

        * step 1 : creates empty folders to save the run results
        * step 2 : reads ``Theta_0`` if specified (as the MLE or MAP)

        Parameters
        ----------
        dict_posteriors : dict[str, Posterior]
            dictionary of posterior distributions
        path_raw : str
            path to the folders where the ``.hdf5`` files are to be stored
        N_runs : int
            number of independent Markov chains to run per posterior distribution
        scaler : Scaler
            contains the transformation of the Theta values from their natural space to their scaled space (in which the sampling happens)
        start_from : Optional[str]
            point at which the inversion will start, must be in [None, "MLE", "MAP"]. For None, a random value is drawn uniformly in the scaled hypercube.
        path_csv_mle : Optional[str]
            path to the csv file containing the already estimated MLE
        path_csv_map : Optional[str]
            path to the csv file containing the already estimated MAP

        Returns
        -------
        Optional[np.ndarray]
            starting point of the  (in scaled space) inversion, ``Theta_0``, if specified. Otherwise ``None``.
        """
        # step 1 : create empty folders to save the run results
        for seed in range(N_runs):
            for model_name in list(dict_posteriors.keys()):
                folder_path = f"{path_raw}/{model_name}/mcmc_{seed}"

                if not os.path.isdir(folder_path):
                    os.mkdir(folder_path)

        # step 2 : read Theta_0 if needed
        assert start_from in ["MLE", "MAP", None]
        model_name = list(dict_posteriors.keys())[0]

        if start_from == "MLE":
            assert path_csv_mle is not None
            Theta_0, _ = ResultsExtractorOptimMLE.read_estimator(
                path_csv_mle,
                model_name,
            )
            Theta_0 = scaler.from_lin_to_scaled(Theta_0)

        elif start_from == "MAP":
            assert path_csv_map is not None
            Theta_0, _ = ResultsExtractorOptimMAP.read_estimator(
                path_csv_map,
                model_name,
            )
            Theta_0 = scaler.from_lin_to_scaled(Theta_0)

        else:
            Theta_0 = None

        return Theta_0

    def run(
        self,
        dict_posteriors: dict[str, Posterior],
        sampler_: Sampler,
        saver_: Saver,
        N_runs: int,
        max_iter: int,
        T_BI: int,
        path_raw: str,
        Theta_0: Optional[np.ndarray] = None,
        # sample_regu_weights: bool = True,
        # T_BI_reguweights: Optional[int] = None,
        #
        regu_spatial_N0: Union[int, float] = np.infty,
        regu_spatial_scale: Optional[float] = 1.0,
        regu_spatial_vmin: Optional[float] = 1e-8,
        regu_spatial_vmax: Optional[float] = 1e8,
        #
        freq_save: int = 1,
        can_run_in_parallel: bool = True,
    ) -> None:
        r"""runs the inversion

        Parameters
        ----------
        dict_posteriors : dict[str, Posterior]
            dictionary of posterior distributions
        sampler_ : Sampler
            sampler to be used to generate the Markov chain(s)
        saver_ : Saver
            object responsible for progressively saving the Markov chain data during the run
        N_runs : int
            number of independent Markov chains to run per posterior distribution
        max_iter : int
            total duration of a Markov chain
        T_BI : int
            duration of the `Burn-in` phase
        path_raw : str
            path to the folders where the ``.hdf5`` files are to be stored
        Theta_0 : Optional[np.ndarray], optional
            starting point, by default None
        regu_spatial_N0 : Union[int, float], optional
            number of iterations defining the initial update phase (for spatial regularization weight optimization). np.infty means that the optimization phase never starts, and that the weight optimization is not applied. by default np.infty
        regu_spatial_scale : Optional[float], optional
            scale parameter involved in the definition of the projected gradient
            step size (for spatial regularization weight optimization). by default 1.0
        regu_spatial_vmin : Optional[float], optional
            lower limit of the admissible interval (for spatial regularization weight optimization), by default 1e-8
        regu_spatial_vmax : Optional[float], optional
            upper limit of the admissible interval (for spatial regularization weight optimization), by default 1e8
        freq_save : int, optional
            frequency of saved iterates during the run (1 means that every iteration is saved), by default 1
        can_run_in_parallel : bool, optional
            wether the inversion can be run in parallel (may cause difficulties for forward maps based on neural networks run on GPU), by default True
        """
        global _run_one_simulation_mcmc_all_pixels

        def _run_one_simulation_mcmc_all_pixels(dict_input: dict) -> dict:
            model_name = dict_input["model_name"]
            seed = dict_input["seed"]

            folder_path = f"{path_raw}/{model_name}/mcmc_{seed}"
            saver_seed = copy.deepcopy(saver_)
            saver_seed.set_results_path(folder_path)

            sampler_seed = copy.deepcopy(sampler_)

            tps0 = time.time()

            sampler_seed.sample(
                dict_posteriors[model_name],
                saver=saver_seed,
                max_iter=max_iter,
                Theta_0=Theta_0,
                #
                # sample_regu_weights=sample_regu_weights,
                # T_BI_reguweights=T_BI_reguweights,
                regu_spatial_N0=regu_spatial_N0,
                regu_spatial_scale=regu_spatial_scale,
                regu_spatial_vmin=regu_spatial_vmin,
                regu_spatial_vmax=regu_spatial_vmax,
                #
                T_BI=T_BI,
            )

            # return input dict with duration information
            total_duration = time.time() - tps0

            dict_output = {
                "seed": seed,
                "model_name": model_name,
                "total_duration": total_duration,
            }
            return dict_output

        # *
        print("starting sampling")
        list_params = [
            {"seed": seed, "model_name": model_name}
            for seed in range(N_runs)
            for model_name in list(dict_posteriors.keys())
        ]
        if can_run_in_parallel:
            with ProcessPoolExecutor(
                max_workers=self.max_workers, mp_context=mp.get_context("fork")
            ) as p:
                list_simulations_durations = list(
                    p.map(_run_one_simulation_mcmc_all_pixels, list_params)
                )
        else:
            list_simulations_durations = [
                _run_one_simulation_mcmc_all_pixels(params) for params in list_params
            ]

        # save run durations
        df_results_sampling = pd.DataFrame(list_simulations_durations)
        filename = f"{self.path_data_csv_out}/durations_MCMC.csv"
        df_results_sampling.to_csv(filename)
        print("sampling done\n")
        return

    def main(
        self,
        dict_posteriors: dict[str, Posterior],
        sampler_: Sampler,
        saver_: Saver,
        scaler: Scaler,
        N_runs: int,
        max_iter: int,
        T_BI: int,
        path_raw: str,
        path_csv_mle: Optional[str],
        path_csv_map: Optional[str],
        start_from: Optional[str],
        # sample_regu_weights: bool = True,
        # T_BI_reguweights: Optional[int] = None,
        #
        regu_spatial_N0: Union[int, float] = np.infty,
        regu_spatial_scale: Optional[float] = 1.0,
        regu_spatial_vmin: Optional[float] = 1e-8,
        regu_spatial_vmax: Optional[float] = 1e8,
        #
        freq_save: int = 1,
        can_run_in_parallel: bool = True,
    ) -> None:
        r"""sequentially calls ``prepare_run`` and ``run``

        Parameters
        ----------
        dict_posteriors : dict[str, Posterior]
            dictionary of posterior distributions
        sampler_ : Sampler
            sampler to be used to generate the Markov chain(s)
        saver_ : Saver
            object responsible for progressively saving the Markov chain data during the run
        scaler : Scaler
            contains the transformation of the Theta values from their natural space to their scaled space (in which the sampling happens)
        N_runs : int
            number of independent Markov chains to run per posterior distribution
        max_iter : int
            total duration of a Markov chain
        T_BI : int
            duration of the `Burn-in` phase
        path_raw : str
            path to the folders where the ``.hdf5`` files are to be stored
        path_csv_mle : Optional[str]
            path to the csv file containing the already estimated MLE
        path_csv_map : Optional[str]
            path to the csv file containing the already estimated MAP
        start_from : Optional[str]
            point at which the inversion will start, must be in [None, "MLE", "MAP"]. For None, a random value is drawn uniformly in the scaled hypercube.
        regu_spatial_N0 : Union[int, float], optional
            number of iterations defining the initial update phase (for spatial regularization weight optimization). np.infty means that the optimization phase never starts, and that the weight optimization is not applied. By default np.infty
        regu_spatial_scale : Optional[float], optional
            scale parameter involved in the definition of the projected gradient
            step size (for spatial regularization weight optimization). by default 1.0
        regu_spatial_vmin : Optional[float], optional
            lower limit of the admissible interval (for spatial regularization weight optimization), by default 1e-8
        regu_spatial_vmax : Optional[float], optional
            upper limit of the admissible interval (for spatial regularization weight optimization), by default 1e8
        freq_save : int, optional
            frequency of saved iterates during the run (1 means that every iteration is saved), by default 1
        can_run_in_parallel : bool, optional
            wether the inversion can be run in parallel (may cause difficulties for forward maps based on neural networks run on GPU), by default True
        """
        Theta_0 = self.prepare_run(
            dict_posteriors,
            path_raw,
            N_runs,
            scaler,
            start_from=start_from,
            path_csv_mle=path_csv_mle,
            path_csv_map=path_csv_map,
        )
        self.run(
            dict_posteriors,
            sampler_,
            saver_,
            N_runs=N_runs,
            max_iter=max_iter,
            T_BI=T_BI,
            path_raw=path_raw,
            Theta_0=Theta_0,
            #
            regu_spatial_N0=regu_spatial_N0,
            regu_spatial_scale=regu_spatial_scale,
            regu_spatial_vmin=regu_spatial_vmin,
            regu_spatial_vmax=regu_spatial_vmax,
            #
            freq_save=freq_save,
            can_run_in_parallel=can_run_in_parallel,
        )
        return



import abc
import os
from typing import Dict, Optional, Tuple

import numpy as np

from beetroots.modelling.forward_maps.abstract_base import ForwardMap
from beetroots.space_transform.abstract_transform import Scaler


class SimulationForwardMap(abc.ABC):
    r"""abstract class for to set up the forward map for an inversion of astrophysical data"""

    MODELS_PATH = os.path.abspath(
        f"{os.path.dirname(os.path.abspath(__file__))}/../../../../data/models"
    )
    r"""str: path to the folder containing all the already defined and saved models (i.e., polynomials or neural networks)"""

    @abc.abstractmethod
    def setup_forward_map(self, **kwargs) -> Tuple[Scaler, ForwardMap]:
        """sets up the forward map and the scaler

        Returns
        -------
        Scaler
            contains the transformation of the Theta values from their natural space to their scaled space (used by the forward map and in which the sampling happens) and its inverse
        ForwardMap
            forward map to be used in the inversion
        """
        pass

    def scale_dict_fixed_params(
        self, scaler: Scaler, dict_fixed_params: Dict[str, Optional[float]]
    ) -> Dict[str, Optional[float]]:
        r"""transforms the fixed parameters from their value in their natural spae to their value in the space in which they are to be sampled.

        Note
        ----
        since the scaler does not accept None or np.NaN, this transformation really needed a dedicated (though simple) method

        Parameters
        ----------
        scaler : Scaler
            contains the transformation of the physical parameters values from their natural space to their scaled space (in which the sampling happens) and its inverse
        dict_fixed_params : Dict[str, Optional[float]]
            contains the value of the fixed parameters in the physical parameter natural space. For example, {"kappa":None, "Pth":None, "G0":None, "AV":None, "angle":0.} (the values with None indicate that the parameter is not fixed, so here, only the angle is set, to 0 degree).

        Returns
        -------
        Dict[str, Optional[float]]
            contains the value of the fixed parameters in the space in which the physical parameter are to be sampled
        """
        arr_fixed_values = np.ones((1, len(dict_fixed_params)))
        for i, name in enumerate(dict_fixed_params.keys()):
            if dict_fixed_params[name] is not None:
                arr_fixed_values[0, i] = dict_fixed_params[name] * 1

        arr_fixed_values_scaled = scaler.from_lin_to_scaled(arr_fixed_values)

        dict_fixed_params_scaled = {
            name: arr_fixed_values_scaled[0, i]
            if dict_fixed_params[name] is not None
            else None
            for i, name in enumerate(dict_fixed_params.keys())
        }
        return dict_fixed_params_scaled



from typing import Dict, List, Union

import numpy as np
from scipy.stats import gaussian_kde
from tqdm.auto import tqdm

from beetroots.approx_optim.abstract_approx_optim import ApproxParamsOptim
from beetroots.approx_optim.approach_type.bo import BayesianOptimizationApproach
from beetroots.approx_optim.forward_map.nn import ApproxOptimNN


class ApproxParamsOptimNNBO(
    ApproxParamsOptim, ApproxOptimNN, BayesianOptimizationApproach
):
    r"""class that performs likelihood parameter optimization using Bayesian optimization for a neural network forward map"""

    def main(
        self,
        dict_forward_model: Dict[str, Union[str, bool, List[bool], List[float]]],
        lower_bounds_lin: Union[np.ndarray, List],
        upper_bounds_lin: Union[np.ndarray, List],
        n_iter: int,
    ):
        r"""main method of the class, sets up the optimization problems and solves them

        Parameters
        ----------
        dict_forward_model : Dict[str, Union[str, bool, List[bool], List[float]]]
            contains the necessary information to load the forward model with the :class:`NeuralNetworkApprox`
        lower_bounds_lin : Union[np.ndarray, List]
            lower bounds on the physical parameters (in linear scale)
        upper_bounds_lin : Union[np.ndarray, List]
            upper bounds on the physical parameters (in linear scale)
        n_iter : int
            number of iterations for the Bayesian optimization
        """

        assert isinstance(dict_forward_model["forward_model_name"], str)
        # assert isinstance(angle, float)

        if isinstance(lower_bounds_lin, list):
            lower_bounds_lin = np.array(lower_bounds_lin)
        if isinstance(upper_bounds_lin, list):
            upper_bounds_lin = np.array(upper_bounds_lin)

        assert isinstance(lower_bounds_lin, np.ndarray)
        assert isinstance(upper_bounds_lin, np.ndarray)
        assert isinstance(n_iter, int) and n_iter > 0

        if isinstance(lower_bounds_lin, list):
            lower_bounds_lin = np.array(lower_bounds_lin)
        if isinstance(lower_bounds_lin, list):
            lower_bounds_lin = np.array(lower_bounds_lin)

        self.list_idx_sampling = [
            i
            for i, v in enumerate(dict_forward_model["fixed_params"].values())
            if v is None
        ]
        r"""List[int]: indices of physical parameters considered as variables for the likelihood parameter adjustment"""

        self.D_sampling = len(self.list_idx_sampling)
        r"""int: number of physical parameters considered as variables for the likelihood parameter adjustment"""

        self.N_samples_theta = self.K**self.D_sampling
        r"""int: number of samples for :math:`\theta` used to build the histogram of :math:`\log_{10} f_\ell(\theta)`"""

        print("starting setup")
        # step 1: set the bounds for the parameters to be adjusted
        (
            log10_f0,
            bounds_a0_low,
            bounds_a0_high,
            bounds_a1_low,
            bounds_a1_high,
        ) = self.setup_params_bounds()

        # step 2: compute the histogram of log f(theta)
        log10_f_Theta = self.compute_log10_f_Theta(
            dict_forward_model,
            lower_bounds_lin,
            upper_bounds_lin,
        )
        log10_f_Theta_low = log10_f_Theta.min(axis=0)  # (L,)
        log10_f_Theta_high = log10_f_Theta.max(axis=0)  # (L,)

        # step 3: set the KDE on log10 f_\ell(theta)
        print(r"starting evaluation of kde of log10 f_\ell(theta)")
        list_log10_f_grid = np.zeros((self.log10_f_grid_size, self.L))
        pdf_kde_log10_f_Theta = np.zeros((self.log10_f_grid_size, self.L))
        for ell in range(self.L):
            list_log10_f_grid[:, ell] = np.linspace(
                log10_f_Theta_low[ell],
                log10_f_Theta_high[ell],
                self.log10_f_grid_size,
            )
            kde_log10_f_Theta = gaussian_kde(log10_f_Theta[:, ell])
            pdf_kde_log10_f_Theta[:, ell] = kde_log10_f_Theta.pdf(
                list_log10_f_grid[:, ell],
            )
        print(r"evaluation of kde of log10 f_\ell(theta) done")

        print("Starting the optimization")
        for ell in tqdm(range(self.L)):
            print(f"starting line {ell} ({self.list_lines[ell]})")
            self.plot_hist_log10_f_Theta(
                log10_f_Theta[:, ell],
                log10_f_Theta_low[ell],
                log10_f_Theta_high[ell],
                list_log10_f_grid[:, ell],
                pdf_kde_log10_f_Theta[:, ell],
                ell,
            )

            for n in range(self.N):
                pbounds = {
                    "a0": (bounds_a0_low[n, ell], bounds_a0_high[n, ell]),
                    "a1": (bounds_a1_low, bounds_a1_high),
                }
                self.save_setup_to_json(n, ell, pbounds)
                self.optimization(
                    first_points=[(log10_f0[n, ell], 0.5)],
                    init_points=3,
                    n_iter=n_iter,  # 50
                    list_log10_f_grid=list_log10_f_grid[:, ell],
                    pdf_kde_log10_f_Theta=pdf_kde_log10_f_Theta[:, ell],
                    pbounds=pbounds,
                    sigma_a_val=self.sigma_a[n, ell],
                    sigma_m_val=self.sigma_m[n, ell],
                    n=n,
                    ell=ell,
                )

        df_best = self.extract_optimal_params()
        df_best = df_best.set_index(["n", "ell"])

        for n in range(self.N):
            for ell in range(self.L):
                best_point = df_best.loc[(n, ell), ["a0_best", "a1_best"]].values
                self.plot_hist_log10_f_Theta_with_optim_results(
                    log10_f_Theta=log10_f_Theta[:, ell],
                    log10_f_Theta_low=log10_f_Theta_low[ell],
                    log10_f_Theta_high=log10_f_Theta_high[ell],
                    list_log10_f_grid=list_log10_f_grid[:, ell],
                    pdf_kde_log10_f_Theta=pdf_kde_log10_f_Theta[:, ell],
                    n=n,
                    ell=ell,
                    best_point=best_point,
                )

        self.plots_postprocessing(
            bounds_a0_low,
            bounds_a0_high,
            bounds_a1_low,
            bounds_a1_high,
            n_iter,
        )



from abc import ABC, abstractmethod
from typing import Optional, Union

import numpy as np


class Likelihood(ABC):
    r"""Abstract Base Class for a probability distribution on non-countable set"""

    def __init__(
        self,
        forward_map,
        D: int,
        L: int,
        N: int,
        y: np.ndarray,
    ) -> None:
        self.forward_map = forward_map
        self.D = D
        self.L = L
        self.N = N
        self.y = y
        self.hyperparameters = None

        assert y.shape == (N, L)

    def _update_observations(self, y: np.ndarray):
        r"""Update observations :math:`y` to be updated as :math:`y_{\text{new}}`  whenever the likelihood object :math:`p(y \mid f(\theta)` is interpreted as
        a prior on :math:`y` with hyperparameter :math:`\theta`.

        Parameters
        ----------
        y : np.ndarray
            New state of the observations.

        Example
        -------
        # forward_map_eval contains the parameter \theta
        lklhd._update_observations(y_new)
        lklhd.neglog_pdf(forward_map_eval, nll_utils)

        Note
        ----
        Allows evaluation of :math:`-\log p(y_{\text{new}} \mid \theta)`.
        """
        self.y = y

    @abstractmethod
    def neglog_pdf(
        self,
        forward_map_evals: dict,
        nll_utils: dict,
        pixelwise: bool = False,
        full: bool = False,
        idx: Optional[np.ndarray] = None,
    ) -> Union[float, np.ndarray]:
        pass

    @abstractmethod
    def sample_observation_model(
        self, forward_map_evals: dict, rng: Optional[np.random.Generator]
    ) -> np.ndarray:
        pass

    @abstractmethod
    def gradient_neglog_pdf(
        self,
        forward_map_evals: dict[str, np.ndarray],
        nll_utils: dict[str, np.ndarray],
    ) -> np.ndarray:
        pass

    def neglog_pdf_candidates(
        self,
        candidates: np.ndarray,
        idx: np.ndarray,
        Theta_t: Optional[np.ndarray] = None,
        return_forward_map_evals: bool = False,
    ) -> np.ndarray:
        assert len(candidates.shape) == 2 and candidates.shape[1] == self.D
        assert isinstance(idx, np.ndarray)
        assert np.all(0 <= idx)
        assert np.all(idx <= self.N - 1)

        N_candidates = candidates.shape[0]
        n_pix = idx.size

        forward_map_evals = self.evaluate_all_forward_map(
            candidates, compute_derivatives=False, compute_derivatives_2nd_order=False
        )
        nll_utils = self.evaluate_all_nll_utils(
            forward_map_evals,
            idx=idx,
            compute_derivatives=False,
            compute_derivatives_2nd_order=False,
        )

        nll_candidates = self.neglog_pdf(
            forward_map_evals,
            nll_utils,
            pixelwise=True,
            idx=idx,
        )  # (N_candidates,)
        assert isinstance(nll_candidates, np.ndarray)
        assert nll_candidates.shape == (N_candidates,)

        if return_forward_map_evals:
            return nll_candidates, forward_map_evals

        else:
            return nll_candidates

    @abstractmethod
    def hess_diag_neglog_pdf(
        self, forward_map_evals: dict, nll_utils: dict
    ) -> np.ndarray:
        pass

    def evaluate_all_forward_map(
        self,
        Theta: np.ndarray,
        compute_derivatives: bool,
        compute_derivatives_2nd_order: bool,
    ) -> dict[str, Union[float, np.ndarray]]:
        assert len(Theta.shape) == 2 and Theta.shape[1] == self.D
        forward_map_evals = self.forward_map.compute_all(
            Theta, True, True, compute_derivatives, compute_derivatives_2nd_order
        )
        return forward_map_evals

    @abstractmethod
    def evaluate_all_nll_utils(
        self,
        forward_map_evals: dict[str, Union[float, np.ndarray]],
        idx: Optional[np.ndarray],
        compute_derivatives: bool,
        compute_derivatives_2nd_order: bool,
    ) -> dict[str, Union[float, np.ndarray]]:
        pass



import os
from typing import List, Tuple

import h5py
import matplotlib.pyplot as plt
import numpy as np

from beetroots.inversion.results.utils.abstract_util import ResultsUtil

# TODO: for now, assumes that the Sampler has a MTM kernel and a PMALA kernel
# -> generalize ?


class ResultsKernels(ResultsUtil):

    __slots__ = (
        "model_name",
        "chain_type",
        "path_img",
        "N_run",
        "effective_T",
    )

    def __init__(
        self,
        model_name: str,
        chain_type: str,
        path_img: str,
        N_run: int,
        T: int,
        freq_save: int,
    ):
        assert chain_type in ["mcmc", "optim_map", "optim_mle"]

        self.model_name = model_name
        self.chain_type = chain_type
        self.path_img = path_img

        self.N_run = N_run
        self.effective_T = T // freq_save

    def read_data(
        self,
        list_chains_folders: List[str],
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        list_type = np.zeros((self.N_run, self.effective_T))
        list_accepted = np.zeros((self.N_run, self.effective_T))
        list_log_proba = np.zeros((self.N_run, self.effective_T))

        for seed, mc_path in enumerate(list_chains_folders):
            with h5py.File(mc_path, "r") as f:
                list_type[seed] = np.array(f["list_type_t"])
                list_accepted[seed] = np.array(f["list_accepted_t"])
                list_log_proba[seed] = np.array(f["list_log_proba_accept_t"])

        return list_type, list_accepted, list_log_proba

    def create_folders(self) -> Tuple[str, str]:
        folder_path_inter = f"{self.path_img}/accepted_freq"
        folder_path_inter2 = f"{folder_path_inter}/{self.chain_type}"
        folder_path_accept_freq = f"{folder_path_inter2}/{self.model_name}"

        for path_ in [
            folder_path_inter,
            folder_path_inter2,
            folder_path_accept_freq,
        ]:
            if not os.path.isdir(path_):
                os.mkdir(path_)

        folder_path_inter = f"{self.path_img}/log_proba_accept"
        folder_path_inter2 = f"{folder_path_inter}/{self.chain_type}"
        folder_path_log_p_accept = f"{folder_path_inter2}/{self.model_name}"
        for path_ in [
            folder_path_inter,
            folder_path_inter2,
            folder_path_log_p_accept,
        ]:
            if not os.path.isdir(path_):
                os.mkdir(path_)

        return (folder_path_accept_freq, folder_path_log_p_accept)

    def plot_accept_freq(
        self,
        folder_path: str,
        list_type: np.ndarray,
        list_accepted: np.ndarray,
    ) -> None:
        # mobile mean size
        k_mm_mtm = 20  # MTM
        k_mm_mala = 20  # P-MALA

        print("starting plot of accepted frequencies")

        # * MTM kernel
        for seed in range(self.N_run):
            idx_mtm = list_type[seed] == 0
            accepted_mtm = list_accepted[seed, idx_mtm]

            if accepted_mtm.size > k_mm_mtm:
                accepted_mtm_smooth = np.convolve(
                    accepted_mtm,
                    np.ones(k_mm_mtm) / k_mm_mtm,
                    mode="valid",
                )

                plt.figure(figsize=(8, 6))
                nan_mean = 100 * np.nanmean(accepted_mtm)
                plt.title(f"MTM : {nan_mean:.2f} % accepted")
                plt.plot(accepted_mtm_smooth, label="mobile mean")
                plt.grid()
                plt.legend()
                plt.xticks(rotation=45)
                # plt.tight_layout()

                filename = f"{folder_path}/freq_accept_seed{seed}_MTM.PNG"
                plt.savefig(filename, bbox_inches="tight")
                plt.close()

        # * PMALA
        for seed in range(self.N_run):
            idx_pmala = list_type[seed] == 1
            accepted_pmala = list_accepted[seed, idx_pmala]

            if accepted_pmala.size > k_mm_mala:
                accepted_pmala_smooth = np.convolve(
                    accepted_pmala,
                    np.ones(k_mm_mala) / k_mm_mala,
                    mode="valid",
                )

                plt.figure(figsize=(8, 6))
                nan_mean = 100 * np.nanmean(accepted_pmala)
                plt.title(f"PMALA : {nan_mean:.2f} % accepted")
                plt.plot(accepted_pmala_smooth, label="mobile mean")
                plt.grid()
                plt.legend()
                plt.xticks(rotation=45)
                # plt.tight_layout()

                filename = f"{folder_path}/freq_accept_seed{seed}_PMALA.PNG"
                plt.savefig(filename, bbox_inches="tight")
                plt.close()

        print("plots of accepted frequencies done")
        return

    def plot_log_proba_accept(
        self,
        folder_path: str,
        list_type: np.ndarray,
        list_log_proba: np.ndarray,
    ) -> None:
        """plots log proba accept per kernel"""

        # mobile mean size
        k_mm_mtm = 20  # MTM
        k_mm_mala = 20  # P-MALA

        print("starting plot of log proba accept")

        # * MTM
        for seed in range(self.N_run):
            idx_mtm = list_type[seed] == 0
            list_log_proba_mtm = list_log_proba[seed, idx_mtm]

            if list_log_proba_mtm.size > k_mm_mtm:
                list_log_proba_mtm_smooth = np.convolve(
                    list_log_proba_mtm,
                    np.ones(k_mm_mtm) / k_mm_mtm,
                    mode="valid",
                )

                plt.figure(figsize=(8, 6))

                nan_mean = np.nanmean(list_log_proba_mtm)
                nan_median = np.nanmedian(list_log_proba_mtm)
                title = f"MTM: log proba accept avg: {nan_mean:.3e},"
                title += f"median: {nan_median:.3e}"
                plt.title(title)
                plt.plot(list_log_proba_mtm_smooth, label="mobile mean")
                # plt.axvline(self.T_BI, c="k", ls="--", label="T_BI")
                plt.grid()
                plt.legend()
                plt.yscale("symlog")
                plt.xticks(rotation=45)
                # plt.tight_layout()

                filename = f"{folder_path}/log_proba_accept_seed{seed}_MTM.PNG"
                plt.savefig(filename, bbox_inches="tight")
                plt.close()

        # * PMALA
        for seed in range(self.N_run):
            idx_pmala = list_type[seed] == 1
            list_log_proba_pmala = list_log_proba[seed, idx_pmala]

            if list_log_proba_pmala.size > k_mm_mala:
                list_log_proba_pmala_smooth = np.convolve(
                    list_log_proba_pmala,
                    np.ones(k_mm_mala) / k_mm_mala,
                    mode="valid",
                )

                plt.figure(figsize=(8, 6))

                nan_mean = np.nanmean(list_log_proba_mtm)
                nan_median = np.nanmedian(list_log_proba_mtm)
                title = f"PMALA: log proba accept avg: {nan_mean:.3e},"
                title += f"median: {nan_median:.3e}"

                plt.title(title)
                plt.plot(list_log_proba_pmala_smooth, label="mobile mean")
                # plt.axvline(self.T_BI, c="k", ls="--", label="T_BI")
                plt.grid()
                plt.legend()
                plt.yscale("symlog")
                plt.xticks(rotation=45)
                # plt.tight_layout()

                filename = f"{folder_path}/log_proba_accept_"
                filename += f"seed{seed}_PMALA.PNG"
                plt.savefig(filename, bbox_inches="tight")
                plt.close()

        print("plots of log proba accept done")
        return

    def main(self, list_chains_folders: List[str]) -> None:
        list_type, list_accepted, list_log_proba = self.read_data(
            list_chains_folders,
        )
        (folder_accept_freq, folder_log_proba_accept) = self.create_folders()

        self.plot_accept_freq(folder_accept_freq, list_type, list_accepted)
        self.plot_log_proba_accept(
            folder_log_proba_accept,
            list_type,
            list_log_proba,
        )



import multiprocessing as mp
from concurrent.futures import ProcessPoolExecutor
from typing import Dict, Union

import numpy as np
import pandas as pd
from scipy.integrate import simpson
from scipy.stats import ks_1samp
from tqdm.auto import tqdm

P_lambda = np.poly1d(np.array([-6.0, 15.0, -10.0, 0.0, 0.0, 1.0]))
SQRT_2_PI = np.sqrt(2 * np.pi)
LOG_10 = np.log(10)


def neg_log_pdf_add(
    y: Union[float, np.ndarray],
    m_a: Union[float, np.ndarray],
    s_a: Union[float, np.ndarray],
) -> Union[float, np.ndarray]:
    """evaluates the negative log pdf of a Gaussian distribution

    Parameters
    ----------
    y : Union[float, np.ndarray]
        points at which the pdf is to be evaluated
    m_a : Union[float, np.ndarray]
        bias of the Gaussian distribution
    s_a : Union[float, np.ndarray]
        standard deviation of the Gaussian distribution

    Returns
    -------
    Union[float, np.ndarray]
        negative log pdf
    """
    return 0.5 * ((y - m_a) / s_a) ** 2 + np.log(SQRT_2_PI * s_a)


def neg_log_pdf_mult(y, m_m, s_m):
    nll_mu = np.where(
        y > 0,
        0.5 * ((np.log(y) - m_m) / s_m) ** 2 + np.log(y * SQRT_2_PI * s_m),
        np.infty,
    )
    return nll_mu


def compute_lambda(
    a0: float, a1: float, f_Theta_true: Union[float, np.ndarray]
) -> Union[float, np.ndarray]:
    r"""careful: a0 and a1 are in log, base 10, and computations require natural base

    Parameters
    ----------
    a0 : float
        center of mixing interval
    a1 : float
        radius of mixing interval
    f_Theta_true : Union[float, np.ndarray]
        value of true :math:`f(\theta)`

    Returns
    -------
    lambda: Union[float, np.ndarray]
        mixing parameter
    """
    log_fm1 = (a0 - a1) * LOG_10
    log_fp1 = (a0 + a1) * LOG_10

    lambda_ = np.where(
        np.log(f_Theta_true) <= log_fm1,
        1,
        np.where(
            np.log(f_Theta_true) >= log_fp1,
            0,
            P_lambda((np.log(f_Theta_true) - log_fm1) / (log_fp1 - log_fm1)),
        ),
    )
    return lambda_


def pdf_my_approx(y, lambda_, m_a, s_a, m_m, s_m, f_Theta_true):
    neg_log_pdf_add_arr = neg_log_pdf_add(y, +m_a + f_Theta_true, s_a)
    neg_log_pdf_mult_arr = np.nan_to_num(
        neg_log_pdf_mult(y, +m_m + np.log(f_Theta_true), s_m)
    )

    log_res = neg_log_pdf_add_arr * lambda_ + neg_log_pdf_mult_arr * (1 - lambda_)
    return np.exp(-log_res)


def evaluate_pdf(y, a0, a1, f_Theta_true, sigma_a, sigma_m):
    lambda_ = compute_lambda(a0, a1, f_Theta_true)

    m_a = (np.exp(sigma_m**2 / 2) - 1) * f_Theta_true
    s_a = np.sqrt(
        f_Theta_true**2 * np.exp(sigma_m**2) * (np.exp(sigma_m**2) - 1)
        + sigma_a**2
    )

    combination = (sigma_a / f_Theta_true) * np.exp(-(sigma_m**2))
    m_m = -0.5 * np.log(1 + combination**2)
    s_m = np.sqrt(sigma_m**2 - 2 * m_m)

    my_pdf = pdf_my_approx(y, lambda_, m_a, s_a, m_m, s_m, f_Theta_true)
    return my_pdf


def evaluate_cdf(y, a0, a1, f_Theta_true, sigma_a, sigma_m):
    idx = np.argsort(y)
    y = y[idx]
    dy = np.diff(y, prepend=0)
    dy[0] = 0

    my_pdf = evaluate_pdf(y, a0, a1, f_Theta_true, sigma_a, sigma_m)
    my_pdf = my_pdf[idx]

    my_cdf = (my_pdf * dy).cumsum()
    my_cdf /= my_cdf[-1]  # normalize
    return my_cdf


def _estimate_one_dks(dict_input: Dict):
    a0 = dict_input["a0"]
    a1 = dict_input["a1"]
    i = dict_input["i"]
    f_Theta_true = dict_input["f_Theta_true"]
    sigma_a = dict_input["sigma_a"]
    sigma_m = dict_input["sigma_m"]
    N_samples_y = dict_input["N_samples_y"]

    samples_a = np.random.normal(scale=sigma_a, size=N_samples_y)
    samples_m = np.random.lognormal(sigma=sigma_m, size=N_samples_y)

    samples_true = samples_m * f_Theta_true + samples_a
    samples_true.sort()

    dks, _ = ks_1samp(
        samples_true,
        evaluate_cdf,
        args=(a0, a1, f_Theta_true, sigma_a, sigma_m),
    )
    dict_output = {"i": i, "dks": dks}
    return dict_output


def estimate_avg_dks_full_bo(
    a0: float,
    a1: float,
    list_log10_f_grid: np.ndarray,
    pdf_kde_log10_f_Theta: np.ndarray,
    sigma_a: float,
    sigma_m: float,
    N_samples_y: int,
    max_workers: int,
) -> float:
    list_params = [
        {
            "a0": a0,
            "a1": a1,
            "i": i,
            "f_Theta_true": f_Theta_true,
            "sigma_a": sigma_a,
            "sigma_m": sigma_m,
            "N_samples_y": N_samples_y,
        }
        for i, f_Theta_true in enumerate(np.exp(list_log10_f_grid * np.log(10.0)))
    ]
    with ProcessPoolExecutor(
        max_workers=max_workers, mp_context=mp.get_context("fork")
    ) as p:
        list_results = list(
            tqdm(p.map(_estimate_one_dks, list_params), total=len(list_params))
        )

    df_results = pd.DataFrame.from_records(list_results)
    df_results = df_results.sort_values("i")
    list_dks = df_results["dks"].values
    assert len(df_results) == len(list_params)
    assert list_dks.shape == pdf_kde_log10_f_Theta.shape

    # ! average of Dks weighted with pdf of log10_f_Theta
    # ! integrated between min and max of log10_f_Theta
    avg_dist = simpson(
        y=pdf_kde_log10_f_Theta * list_dks,
        x=list_log10_f_grid,
    )
    # print(avg_dist)

    return -np.log10(avg_dist)



r"""Former version of the ``latex-ism-emission-lines`` package by L. Einig available at  https://github.com/einigl/latex-ism-emission-lines/tree/main.
"""

import re
from typing import Optional, Sequence, Tuple, Union

__all__ = ["lines_to_latex"]


# Molecular names with underscore
_underscore_prefixes = {
    "13c_o": "13co",
    "c_18o": "c-18o",
    "13c_18o": "13c-18o",
    "h2_18o": "h2-18o",
    "c_c3h2": "c-c3h2",
}

# Molecular names to latex
_molecules_to_latex = {
    "h": "H",
    "h2": "H_2",
    "hd": "HD",
    "co": "CO",
    "13co": "^{13}CO",
    "c-18o": "C^{18}O",
    "13c-18o": "^{13}C^{18}O",
    "c": "C",
    "n": "N",
    "o": "O",
    "s": "S",
    "si": "Si",
    "cs": "CS",
    "cn": "CN",
    "hcn": "HCN",
    "hnc": "HNC",
    "oh": "OH",
    "h2o": "H_2O",
    "h2-18o": "H_2^{18}O",
    "c2h": "C_2H",
    "c-c3h2": "c-C_3H_2",
    "so": "SO",
    "cp": "C^+",
    "sp": "S^+",
    "hcop": "HCO^+",
    "chp": "CH^+",
    "ohp": "OH^+",
    "shp": "SH^+",
}

# Energy to LaTeX
_energy_to_latex = {
    "j": "J",
    "v": "\\nu",
    "f": "f",
    "n": "n",
    "ka": "k_a",
    "kc": "k_c",
}


def _transition(
    name: Optional[str], high_lvl: str, low_lvl: str
) -> Tuple[Union[str, Tuple[str, str]], bool]:
    """
    Returns a LaTeX string representing a non electronic transition.

    Parameters
    ----------
    name : str
        Energy name.
    high_lvl : str
        Higher energy level.
    low_lvl : str
        Lower energy level. Can be the same as `high_lvl`.

    Returns
    -------
    str or tuple of str
        If it is a transition: higher energy level and lower energy level. Else: energy level.
    bool
        True if it is a transition, else False
    """
    if name:
        if name in _energy_to_latex:
            name_latex = _energy_to_latex[name]
        else:
            name_latex = name
        if high_lvl == low_lvl:
            return "${}={}$".format(name_latex, low_lvl), False
        return (
            "${}={}$".format(name_latex, high_lvl),
            "${}={}$".format(name_latex, low_lvl),
        ), True
    if high_lvl == low_lvl:
        return "", False
    return ("${}$".format(high_lvl), "${}$".format(low_lvl)), True


def _eltransition(high: str, low: str) -> Tuple[Union[str, Tuple[str, str]], bool]:
    """
    Returns a LaTeX string representing an electronic transition.

    Parameters
    ----------
    high : str
        Higher energy electronic configuration.
    low : str
        Lower energy electronic configuration. Can be the same as `high`.

    Returns
    -------
    str or tuple of str
        If it is a transition: higher energy electronic configuration and lower energy configuration. Else: energy electronic configuration.
    bool
        True if it is a transition, else False
    """
    if high == low:
        return "${}$".format(high), False
    return ("${}$".format(high), "${}$".format(low)), True


def _sort_transitions(
    names: Sequence[str], high_lvls: Sequence[int], low_lvls: Sequence[int]
) -> str:
    """
    Returns a LaTeX string representing the energy transitions.
    This string first display the constant energy levels and then the energy transitions.

    Parameters
    ----------
    names : Sequence of str
        Energies names.
    high_lvls : Sequence of int
        Sequence of higher level for each energy.
    low_lvls : Sequence of int.
        Sequence of lower level for each energy.

    Returns
    -------
    str
        String representing first the constant energy levels and then the energy transitions.
    """
    if len(high_lvls) != len(names) or len(low_lvls) != len(names):
        raise ValueError("names, high_lvls and low_lvls must have the same length")

    if len(names) == 0:
        return ""
    if len(names) == 1:
        return _transition(None, high_lvls[0], low_lvls[0])

    descr_0, descr_1a, descr_1b = "", "", ""
    for name, high, low in zip(names, high_lvls, low_lvls):
        if name == "el":
            descr, istrans = _eltransition(high, low)
        else:
            descr, istrans = _transition(name, high, low)
        if istrans:
            descr_1a += descr[0] + ", "
            descr_1b += descr[1] + ", "
        else:
            descr_0 += descr + " "
    return "{} ({} $\\to$ {})".format(descr_0.strip(), descr_1a[:-2], descr_1b[:-2])


def lines_to_latex(line_name: str) -> str:
    """
    Returns a well displayed version of the formatted line ``line_name``.

    Parameters
    ----------
    line_name : str
        Formatted line.

    Returns
    -------
    str
        LaTeX string representing ``line_name``.
    """

    # Check if the input is in the good format
    if not "_" in line_name:
        return line_name
    line_name = line_name.lower().strip()

    # # Replace the underscore in molecular names containing one
    # for pref in _underscore_prefixes:
    #     if pref in line_name:
    #         line_name = line_name.replace(pref, "-", 1)
    #         break

    # Replace the underscore in molecular names containing one
    for pref, new_pref in _underscore_prefixes.items():
        if line_name.startswith(pref):
            line_name = line_name.replace(pref, f"{new_pref}", 1)
            break

    # Split the line name in two parts
    prefix, suffix = line_name.split("_", maxsplit=1)

    # Convert the prefix in LaTeX
    if prefix in _molecules_to_latex:
        latex_prefix = "${}$".format(_molecules_to_latex[prefix])
    else:
        latex_prefix = prefix

    # Convert the suffix in LaTeX
    if re.match("\Aj\d*__j\d*\Z", suffix):
        # ??__j{}__j{}
        res = re.search("\Aj(.*?)__j(.*?)\Z", suffix)
        if res:
            latex_suffix = "$(J={} \\to J={})$".format(*res.group(1, 2))
        else:
            raise ValueError("Should never been here")
    elif re.match("\Av\d*_j\d*__v\d*_j\d*\Z", suffix):
        # ??_v{}_j{}__v{}_j{}
        res = re.search("\Av(.*?)_j(.*?)__v(.*?)_j(.*?)\Z", suffix)
        if res:
            names = ["v", "j"]
            high_lvls, low_lvls = res.group(1, 2), res.group(3, 4)
            latex_suffix = _sort_transitions(names, high_lvls, low_lvls)
        else:
            raise ValueError("Should never been here")
    elif re.match("\An\d*_j\d*__n\d*_j\d*\Z", suffix):
        # ??_n{}_j{}__n{}_j{}
        res = re.search("\An(.*?)_j(.*?)__n(.*?)_j(.*?)\Z", suffix)
        if res:
            names = ["n", "j"]
            high_lvls, low_lvls = res.group(1, 2), res.group(3, 4)
            latex_suffix = _sort_transitions(names, high_lvls, low_lvls)
        else:
            raise ValueError("Should never been here")
    elif re.match("\Aj\d*_f\d*__j\d*_f\d*\Z", suffix):
        # ??_j{}_f{}__j{}_f{}
        res = re.search("\Aj(.*?)_f(.*?)__j(.*?)_f(.*?)\Z", suffix)
        if res:
            names = ["j", "f"]
            high_lvls, low_lvls = res.group(1, 2), res.group(3, 4)
            latex_suffix = _sort_transitions(names, high_lvls, low_lvls)
        else:
            raise ValueError("Should never been here")
    elif re.match("\Aj\d*_ka\d*_kc\d*__j\d*_ka\d*_kc\d*\Z", suffix):
        # ??_j{}_ka{}_kc{}__j{}_ka{}_kc{}
        res = re.search("\Aj(.*?)_ka(.*?)_kc(.*?)__j(.*?)_ka(.*?)_kc(.*?)\Z", suffix)
        if res:
            names = ["j", "ka", "kc"]
            high_lvls, low_lvls = res.group(1, 2, 3), res.group(4, 5, 6)
            latex_suffix = _sort_transitions(names, high_lvls, low_lvls)
        else:
            raise ValueError("Should never been here")
    elif re.match("\Ael\d*\w_j\d*__el\d*\w_j\d*\Z", suffix):
        # ??_el{int+char}_j{int}__el{int+char}_j{int}
        res = re.search("\Ael(.*?)_j(.*?)__el(.*?)_j(.*?)\Z", suffix)
        if res:
            names = ["el", "j"]
            high_lvls, low_lvls = res.group(1, 2), res.group(3, 4)
            latex_suffix = _sort_transitions(names, high_lvls, low_lvls)
    elif re.match("\Ael\d*\w_j\d*_2__el\d*\w_j\d*_2\Z", suffix):
        # ??_el{int+char}_j{int}_2__el{int+char}_j{int}_2
        res = re.search("\Ael(.*?)_j(.*?)_2__el(.*?)_j(.*?)_2\Z", suffix)
        if res:
            names = ["el", "j"]
            high_lvls = res.group(1), r"\frac{" + f"{res.group(2)}" + r"}{2}"
            low_lvls = res.group(3), r"\frac{" + f"{res.group(4)}" + r"}{2}"
            latex_suffix = _sort_transitions(names, high_lvls, low_lvls)
    else:
        # Not adressed formats

        # oh_j5_2_pp_fif2__j3_2_pm_fif2

        # c2h_n1d0_j1d5_f1d0__n0d0_j0d5_f1d0

        # cn_n1_j0d5__n0_j0d5
        # ohp_n1_j0_f0d5__n0_j1_f0d5
        # shp_n1_j0_f0d5__n0_j1_f0d5

        latex_suffix = suffix

    out = latex_prefix + " " + latex_suffix
    out = out.replace("  ", " ")
    return out



from typing import List

import numba
import numpy as np

from beetroots.modelling.priors.abstract_prior import PriorProbaDistribution


@numba.jit("float64[:](float64[:,:], float64[:], float64[:], float64)", nopython=True)
def penalty(
    Theta: np.ndarray,
    lower_bounds: np.ndarray,
    upper_bounds: np.ndarray,
    indicator_margin_scale: float,
) -> np.ndarray:
    D = lower_bounds.size

    neglog_p = np.zeros((D,))
    for d in range(D):
        neglog_p[d] = np.sum(
            np.where(
                Theta[:, d] > upper_bounds[d],
                (Theta[:, d] - upper_bounds[d]) / indicator_margin_scale,
                np.where(
                    Theta[:, d] < lower_bounds[d],
                    (lower_bounds[d] - Theta[:, d]) / indicator_margin_scale,
                    0,
                ),
            )
            ** 4
        )
    return neglog_p  # (D,)


@numba.njit()
def penalty_one_pix(
    Theta: np.ndarray,
    lower_bounds: np.ndarray,
    upper_bounds: np.ndarray,
    indicator_margin_scale: float,
) -> np.ndarray:
    neglog_p_full = (
        np.where(
            Theta > np.expand_dims(upper_bounds, 0),
            (Theta - np.expand_dims(upper_bounds, 0)) / indicator_margin_scale,
            np.where(
                Theta < np.expand_dims(lower_bounds, 0),
                (np.expand_dims(lower_bounds, 0) - Theta) / indicator_margin_scale,
                np.zeros_like(Theta),
            ),
        )
        ** 4
    )  # (N_candidates, D)
    return np.sum(neglog_p_full, axis=1)  # (N_candidates,)


@numba.jit("float64[:,:](float64[:,:], float64[:], float64[:], float64)", nopython=True)
def gradient_penalty(
    Theta: np.ndarray,
    lower_bounds: np.ndarray,
    upper_bounds: np.ndarray,
    indicator_margin_scale: float,
) -> np.ndarray:
    D = lower_bounds.size

    g = np.zeros_like(Theta)
    for d in range(D):
        g[:, d] = (
            4
            / indicator_margin_scale**4
            * np.where(
                Theta[:, d] > upper_bounds[d],
                (Theta[:, d] - upper_bounds[d]),
                np.where(
                    Theta[:, d] < lower_bounds[d],
                    (-lower_bounds[d] + Theta[:, d]),
                    0,
                ),
            )
            ** 3
        )
    return g


@numba.jit("float64[:,:](float64[:,:], float64[:], float64[:], float64)", nopython=True)
def hess_diag_penalty(
    Theta: np.ndarray,
    lower_bounds: np.ndarray,
    upper_bounds: np.ndarray,
    indicator_margin_scale: float,
) -> np.ndarray:
    D = lower_bounds.size

    hess_diag = np.zeros_like(Theta)
    for d in range(D):
        hess_diag[:, d] = (
            12
            / indicator_margin_scale**4
            * np.where(
                Theta[:, d] > upper_bounds[d],
                (Theta[:, d] - upper_bounds[d]),
                np.where(
                    Theta[:, d] < lower_bounds[d],
                    (-lower_bounds[d] + Theta[:, d]),
                    0,
                ),
            )
            ** 2
        )
    return hess_diag  # (N, D)


class SmoothIndicatorPrior(PriorProbaDistribution):
    r"""This prior encodes validity intervals :math:`[\underline{\theta}_{d}, \overline{\theta}_{d}]]` the physical parameters :math:`\theta_{nd}` (for :math:`d \in [\![1, D]\!]`).
    The negative log of this prior is

    .. math::

        \forall n, d, \quad \iota^{\Delta}_{[\underline{\theta}_{d}, \overline{\theta}_{d}]}(\theta_{n,d}) = \begin{cases}
        0 \quad \text{ if } \theta_{n,d} \in [\underline{\theta}_{d}, \overline{\theta}_{d}]\\
        \left(\frac{\theta_{n,d} - \underline{\theta}_{d}}{\Delta} \right)^4 \quad \text{ if } \theta_{n,d} < \underline{\theta}_{d}\\
        \left(\frac{\theta_{n,d} - \overline{\theta}_{d}}{\Delta} \right)^4 \quad \text{ if } \theta_{n,d} > \overline{\theta}_{d}
        \end{cases}

    with :math:`\Delta > 0` a margin scaling parameter.
    """

    __slots__ = (
        "D",
        "N",
        "indicator_margin_scale",
        "lower_bounds",
        "upper_bounds",
    )

    def __init__(
        self,
        D: int,
        N: int,
        indicator_margin_scale: float,
        lower_bounds: np.ndarray,
        upper_bounds: np.ndarray,
        list_idx_sampling: List[int],
    ) -> None:
        super().__init__(D, N)
        self.indicator_margin_scale = indicator_margin_scale
        r"""float: scaling parameter :math:`\Delta`"""

        self.lower_bounds_full = lower_bounds
        r"""np.ndarray: validity interval lower bounds of the full set of D physical parameters"""
        self.upper_bounds_full = upper_bounds
        r"""np.ndarray: validity interval upper bounds of the full set of D physical parameters"""

        self.lower_bounds = lower_bounds[list_idx_sampling]
        r"""np.ndarray: validity interval lower bounds of the set of sampled physical parameters"""
        self.upper_bounds = upper_bounds[list_idx_sampling]
        r"""np.ndarray: validity interval upper bounds of the set of sampled physical parameters"""

        assert (
            self.lower_bounds.size == self.D
        ), f"should be {self.D}, is {self.lower_bounds.size}"
        assert (
            self.upper_bounds.size == self.D
        ), f"should be {self.D}, is {self.upper_bounds.size}"

        return

    def neglog_pdf(self, Theta: np.ndarray, pixelwise: bool = False) -> np.ndarray:
        r"""compute the negative log of the prior that approximates the indicator function

        .. math::

            \forall n, d, \quad \iota^{\Delta}_{[\underline{\theta}_{d}, \overline{\theta}_{d}]}(\theta_{n,d}) = \begin{cases}
            0 \quad \text{ if } \theta_{n,d} \in [\underline{\theta}_{d}, \overline{\theta}_{d}]\\
            \left(\frac{\theta_{n,d} - \underline{\theta}_{d}}{\Delta} \right)^4 \quad \text{ if } \theta_{n,d} < \underline{\theta}_{d}\\
            \left(\frac{\theta_{n,d} - \overline{\theta}_{d}}{\Delta} \right)^4 \quad \text{ if } \theta_{n,d} > \overline{\theta}_{d}
            \end{cases}

        Parameters
        ----------
        Theta : np.array of shape (N, D)
            current iterate
        pixelwise : bool, optional
            wether to return an aggregated result per pixel (if True) or per map (if False), by default False

        Returns
        -------
        neglog_p : np.ndarray of shape (D,) or (N,)
            negative log of the smooth indicator prior pdf
        """
        assert len(Theta.shape) == 2 and Theta.shape[1] == self.D
        if pixelwise:
            neglog_p = penalty_one_pix(
                Theta,
                self.lower_bounds,
                self.upper_bounds,
                self.indicator_margin_scale,
            )  # (N,)
        else:
            neglog_p = penalty(
                Theta,
                self.lower_bounds,
                self.upper_bounds,
                self.indicator_margin_scale,
            )  # (D,)
        return neglog_p

    def neglog_pdf_one_pix(self, Theta: np.ndarray) -> np.ndarray:
        r"""compute the negative log of the prior that approximates the indicator function

        .. math::

            \forall n, d, \quad \iota^{\Delta}_{[\underline{\theta}_{d}, \overline{\theta}_{d}]}(\theta_{n,d}) = \begin{cases}
            0 \quad \text{ if } \theta_{n,d} \in [\underline{\theta}_{d}, \overline{\theta}_{d}]\\
            \left(\frac{\theta_{n,d} - \underline{\theta}_{d}}{\Delta} \right)^4 \quad \text{ if } \theta_{n,d} < \underline{\theta}_{d}\\
            \left(\frac{\theta_{n,d} - \overline{\theta}_{d}}{\Delta} \right)^4 \quad \text{ if } \theta_{n,d} > \overline{\theta}_{d}
            \end{cases}

        Parameters
        ----------
        Theta : np.array of shape (N_candidates, D)
            current iterate

        Returns
        -------
        neglog_p : numpy array of shape (N_candidates,)
            negative log of the smooth indicator prior per map
        """
        assert len(Theta.shape) == 2 and Theta.shape[1] == self.D
        neglog_p = penalty_one_pix(
            Theta,
            self.lower_bounds,
            self.upper_bounds,
            self.indicator_margin_scale,
        )  # (N_candidates,)
        return neglog_p

    def gradient_neglog_pdf(self, Theta: np.ndarray) -> np.ndarray:
        r"""gradient of the negative log pdf of the smooth indicator prior

        Parameters
        ----------
        Theta : np.array of shape (N, D)
            current iterate

        Returns
        -------
        g : np.array of shape (N, D)
            gradient
        """
        assert len(Theta.shape) == 2 and Theta.shape[1] == self.D
        grad_ = gradient_penalty(
            Theta,
            self.lower_bounds,
            self.upper_bounds,
            self.indicator_margin_scale,
        )  # (N, D)
        return grad_  # / (self.N * self.D)

    def hess_diag_neglog_pdf(self, Theta: np.ndarray) -> np.ndarray:
        r"""diagonal of the Hessian of the negative log pdf of the smooth indicator prior

        Parameters
        ----------
        Theta : np.array of shape (N, D)
            current iterate

        Returns
        -------
        hess_diag : np.array of shape (N, D)
            [description]
        """
        assert len(Theta.shape) == 2 and Theta.shape[1] == self.D
        hess_diag = hess_diag_penalty(
            Theta,
            self.lower_bounds,
            self.upper_bounds,
            self.indicator_margin_scale,
        )  # (N, D)
        return hess_diag  # / (self.N * self.D)



import copy
import multiprocessing as mp
import os
import time
from concurrent.futures import ProcessPoolExecutor
from typing import Optional

import numpy as np
import pandas as pd

from beetroots.inversion.results.results_optim_map import ResultsExtractorOptimMAP
from beetroots.inversion.results.results_optim_mle import ResultsExtractorOptimMLE
from beetroots.inversion.run.abstract_run import Run
from beetroots.modelling.posterior import Posterior
from beetroots.sampler.abstract_sampler import Sampler
from beetroots.sampler.saver.abstract_saver import Saver
from beetroots.space_transform.abstract_transform import Scaler


class RunOptimMAP(Run):
    r"""class that runs inversions using an optimization approach, considering all pixels / components at once"""

    __slots__ = ("path_data_csv_out", "max_workers")

    def __init__(self, path_data_csv_out: str, max_workers: int):
        r"""

        Parameters
        ----------
        path_data_csv_out : str
            path to the folder where results are to be saved
        max_workers : int
            max number of workers that can be used to run the inversion
        """
        self.path_data_csv_out = path_data_csv_out
        r"""str: path to the folder where results are to be saved"""

        self.max_workers = max_workers
        r"""int: max number of workers that can be used to run the inversion"""

    def prepare_run(
        self,
        dict_posteriors: dict[str, Posterior],
        path_raw: str,
        N_runs: int,
        scaler: Scaler,
        start_from: Optional[str] = None,
        path_csv_mle: Optional[str] = None,
    ) -> Optional[np.ndarray]:
        r"""prepares the run in two ways :

        * step 1 : creates empty folders to save the run results
        * step 2 : reads ``Theta_0`` if specified (as the MLE)

        Parameters
        ----------
        dict_posteriors : dict[str, Posterior]
            dictionary of posterior distributions
        path_raw : str
            path to the folders where the ``.hdf5`` files are to be stored
        N_runs : int
            number of independent optimization runs to run per posterior distribution
        scaler : Scaler
            contains the transformation of the Theta values from their natural space to their scaled space (in which the sampling happens)
        start_from : Optional[str]
            point at which the inversion will start, must be in [None, "MLE"]. For None, a random value is drawn uniformly in the scaled hypercube.
        path_csv_mle : Optional[str]
            path to the csv file containing the already estimated MLE

        Returns
        -------
        Optional[np.ndarray]
            starting point of the  (in scaled space) inversion, ``Theta_0``, if specified. Otherwise ``None``.
        """
        # create empty folders to save the run results
        for seed in range(N_runs):
            for model_name in list(dict_posteriors.keys()):
                folder_path = f"{path_raw}/{model_name}/optim_MAP_{seed}"

                if not os.path.isdir(folder_path):
                    os.mkdir(folder_path)

        # read Theta_0 if needed
        assert start_from in ["MLE", "MAP", None]

        if start_from == "MLE":
            result_extractor = ResultsExtractorOptimMLE(path_csv_mle)
            Theta_0, _ = result_extractor.read_estimator(model_name)
            Theta_0 = scaler.from_lin_to_scaled(Theta_0)

        else:
            Theta_0 = None

        return Theta_0

    def run(
        self,
        dict_posteriors: dict[str, Posterior],
        sampler_: Sampler,
        saver_: Saver,
        N_runs: int,
        max_iter: int,
        path_raw: str,
        Theta_0: Optional[np.ndarray] = None,
        freq_save: int = 1,
        can_run_in_parallel: bool = True,
    ) -> None:
        r"""runs the inversion

        Parameters
        ----------
        dict_posteriors : dict[str, Posterior]
            dictionary of posterior distributions
        sampler_ : Sampler
            optimizer
        saver_ : Saver
            object responsible for progressively saving the optimization run data during the run
        N_runs : int
            number of independent optimization runs to run per posterior distribution
        max_iter : int
            total duration of an optimization run
        path_raw : str
            path to the folders where the ``.hdf5`` files are to be stored
        Theta_0 : Optional[np.ndarray], optional
            starting point, by default None
        freq_save : int, optional
            frequency of saved iterates during the run (1 means that every iteration is saved), by default 1
        can_run_in_parallel : bool, optional
            wether the inversion can be run in parallel (may cause difficulties for forward maps based on neural networks run on GPU), by default True
        """
        global _run_one_simulation_optim_map_all_pixels

        def _run_one_simulation_optim_map_all_pixels(dict_input: dict) -> dict:
            model_name = dict_input["model_name"]
            seed = dict_input["seed"]

            folder_path = f"{path_raw}/{model_name}/optim_MAP_{seed}"
            saver_seed = copy.deepcopy(saver_)
            saver_seed.set_results_path(folder_path)

            sampler_seed = copy.deepcopy(sampler_)

            tps0 = time.time()

            sampler_seed.sample(
                dict_posteriors[model_name],
                saver=saver_seed,
                max_iter=max_iter,
                Theta_0=Theta_0,
            )
            # return input dict with duration information
            dict_output = {
                "seed": seed,
                "model_name": model_name,
                "total_duration": time.time() - tps0,
            }
            return dict_output

        # * global function
        print("starting optimization MAP")
        list_params = [
            {"seed": seed, "model_name": model_name}
            for seed in range(N_runs)
            for model_name in list(dict_posteriors.keys())
        ]
        if can_run_in_parallel:
            with ProcessPoolExecutor(
                max_workers=self.max_workers, mp_context=mp.get_context("fork")
            ) as p:
                list_simulations_durations = list(
                    p.map(_run_one_simulation_optim_map_all_pixels, list_params)
                )
        else:
            # * non parallel version
            list_simulations_durations = []
            for params in list_params:
                duration = _run_one_simulation_optim_map_all_pixels(params)
                list_simulations_durations.append(duration)

        df_results_sampling = pd.DataFrame(list_simulations_durations)
        filename = f"{self.path_data_csv_out}/durations_optim_MAP.csv"
        df_results_sampling.to_csv(filename)
        print("optimization MAP done\n")
        return

    def main(
        self,
        dict_posteriors: dict[str, Posterior],
        sampler_: Sampler,
        saver_: Saver,
        scaler: Scaler,
        N_runs: int,
        max_iter: int,
        path_raw: str,
        path_csv_mle: Optional[str] = None,
        start_from: Optional[str] = None,
        freq_save: int = 1,
        can_run_in_parallel: bool = True,
    ) -> None:
        r"""sequentially calls ``prepare_run`` and ``run``

        Parameters
        ----------
        dict_posteriors : dict[str, Posterior]
            dictionary of posterior distributions
        sampler_ : Sampler
            optimizer
        saver_ : Saver
            object responsible for progressively saving the optimization run data during the run
        scaler : Scaler
            contains the transformation of the Theta values from their natural space to their scaled space (in which the sampling happens)
        N_runs : int
            number of independent optimization runs to run per posterior distribution
        max_iter : int
            total duration of an optimization run
        path_raw : str
            path to the folders where the ``.hdf5`` files are to be stored
        path_csv_mle : Optional[str]
            path to the csv file containing the already estimated MLE
        start_from : Optional[str], optional
            _description_, by default None
        freq_save : int, optional
            frequency of saved iterates during the run (1 means that every iteration is saved), by default 1
        can_run_in_parallel : bool, optional
            wether the inversion can be run in parallel (may cause difficulties for forward maps based on neural networks run on GPU), by default True
        """
        Theta_0 = self.prepare_run(
            dict_posteriors,
            path_raw,
            N_runs,
            scaler,
            path_csv_mle,
            start_from,
        )
        self.run(
            dict_posteriors,
            sampler_,
            saver_,
            N_runs,
            max_iter,
            path_raw,
            Theta_0,
            freq_save,
            can_run_in_parallel,
        )
        return
